import json
import pathlib
from typing import Any

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.models import (
    AddtionalSource,
    CommandVerification,
    ConstVerification,
    DocumentOutputMode,
    FileResult,
    ForcePosixPath,
    ResultStatus,
    Verification,
    VerificationFile,
    VerificationInput,
)


@pytest.mark.parametrize(
    ("attributes", "expected"),
    [
        ({"NOTITLE": "title"}, None),
        ({"TITLE": "TitleValue"}, "TitleValue"),
        ({"document_title": "DocTitle"}, "DocTitle"),
        ({"document_title": "DocTitle", "TITLE": "TitleValue"}, "TitleValue"),
    ],
    ids=repr,
)
def test_title_attribute(attributes: dict[str, Any], expected: str | None):
    assert VerificationFile(document_attributes=attributes).title == expected


@pytest.mark.parametrize(
    ("display", "expected"),
    [
        ("visible", DocumentOutputMode.visible),
        (DocumentOutputMode.visible, DocumentOutputMode.visible),
        ("hidden", DocumentOutputMode.hidden),
        (DocumentOutputMode.hidden, DocumentOutputMode.hidden),
        ("never", DocumentOutputMode.never),
        (DocumentOutputMode.never, DocumentOutputMode.never),
        ("no-index", DocumentOutputMode.no_index),
        ("no_index", DocumentOutputMode.no_index),
        (DocumentOutputMode.no_index, DocumentOutputMode.no_index),
        ("no-match", None),
        (1, None),
    ],
    ids=str,
)
def test_display_attribute(
    display: str | DocumentOutputMode,
    expected: DocumentOutputMode | None,
):
    assert (
        VerificationFile(document_attributes={"DISPLAY": display}).display == expected
    )


@pytest.mark.parametrize(
    ("verification", "expected"),
    [
        (None, []),
        ([], []),
        (
            ConstVerification(status=ResultStatus.FAILURE),
            [ConstVerification(status=ResultStatus.FAILURE)],
        ),
        (
            [
                ConstVerification(status=ResultStatus.FAILURE),
                ConstVerification(status=ResultStatus.SUCCESS),
            ],
            [
                ConstVerification(status=ResultStatus.FAILURE),
                ConstVerification(status=ResultStatus.SUCCESS),
            ],
        ),
    ],
    ids=str,
)
def test_verification_list(
    verification: list[Verification] | Verification | None, expected: list[Verification]
):
    assert VerificationFile(verification=verification).verification_list == expected


test_parse_VerificationFile_params: list[
    tuple[VerificationFile, dict[str, Any], dict[str, Any]]
] = [
    (
        VerificationFile(),
        {},
        {
            "dependencies": set(),
            "document_attributes": {},
            "verification": [],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            dependencies={
                pathlib.Path("bar1"),
                pathlib.Path("bar2"),
            },
        ),
        {
            "dependencies": [
                "bar1",
                "bar2",
            ],
        },
        {
            "dependencies": {
                pathlib.Path("bar1"),
                pathlib.Path("bar2"),
            },
            "document_attributes": {},
            "verification": [],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            document_attributes={
                "title": "Bar bar",
            },
        ),
        {
            "document_attributes": {
                "title": "Bar bar",
            },
        },
        {
            "dependencies": set(),
            "document_attributes": {
                "title": "Bar bar",
            },
            "verification": [],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            verification=[ConstVerification(status=ResultStatus.SUCCESS)],
        ),
        {
            "verification": [
                {
                    "type": "const",
                    "status": "success",
                }
            ],
        },
        {
            "dependencies": set(),
            "document_attributes": {},
            "verification": [
                {
                    "type": "const",
                    "status": "success",
                }
            ],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            verification=[ConstVerification(status=ResultStatus.SUCCESS)],
        ),
        {
            "verification": [
                {
                    "type": "const",
                    "status": "success",
                }
            ],
        },
        {
            "dependencies": set(),
            "document_attributes": {},
            "verification": [
                {
                    "type": "const",
                    "status": "success",
                }
            ],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            additonal_sources=[
                AddtionalSource(name="dummy", path=pathlib.Path("tmp/dummy.sh"))
            ]
        ),
        {
            "additonal_sources": [{"name": "dummy", "path": "tmp/dummy.sh"}],
        },
        {
            "dependencies": set(),
            "document_attributes": {},
            "verification": [],
            "additonal_sources": [
                {"name": "dummy", "path": pathlib.Path("tmp/dummy.sh")}
            ],
        },
    ),
]


@pytest.mark.parametrize(
    ("obj", "raw_dict", "output_dict"),
    test_parse_VerificationFile_params,
)
def test_parse_VerificationFile(
    obj: VerificationFile,
    raw_dict: dict[str, Any],
    output_dict: dict[str, Any],
):
    assert obj == VerificationFile.model_validate(raw_dict)
    assert obj.model_dump(exclude_none=True) == output_dict


test_is_verification_params = [
    (
        VerificationFile(
            verification=[ConstVerification(status=ResultStatus.SUCCESS)],
        ),
        True,
        True,
    ),
    (
        VerificationFile(
            verification=[
                ConstVerification(status=ResultStatus.SUCCESS),
                ConstVerification(status=ResultStatus.FAILURE),
            ],
        ),
        True,
        True,
    ),
    (
        VerificationFile(
            verification=[CommandVerification(command="true")],
        ),
        True,
        False,
    ),
    (
        VerificationFile(
            verification=[
                ConstVerification(status=ResultStatus.SUCCESS),
                CommandVerification(command="true"),
            ],
        ),
        True,
        False,
    ),
    (
        VerificationFile(
            verification=[],
        ),
        False,
        False,
    ),
]


@pytest.mark.parametrize(
    ("obj", "is_verification", "is_lightweight_verification"),
    test_is_verification_params,
)
def test_is_verification(
    obj: VerificationFile,
    is_verification: bool,
    is_lightweight_verification: bool,
):
    assert obj.is_verification() == is_verification
    assert obj.is_lightweight_verification() == is_lightweight_verification


def test_parse_file_relative(testtemp: pathlib.Path, mocker: MockerFixture):
    mocker.patch.object(pathlib.Path, "cwd", return_value=testtemp)
    tmp = testtemp / "verify.json"
    with tmp.open("w") as fp:
        json.dump(
            {
                "files": {
                    (testtemp / "libfile.py").as_posix(): {
                        "dependencies": [
                            (testtemp / "libfile2.py").as_posix(),
                        ]
                    },
                    (testtemp / "libfile2.py").as_posix(): {"dependencies": []},
                    "/foo/other/libfile.py": {"dependencies": []},
                    (testtemp / "test/test.py").as_posix(): {
                        "dependencies": ["/foo/other/libfile.py", "../libfile.py"]
                    },
                }
            },
            fp,
        )
    assert (
        VerificationInput.parse_file_relative(tmp).model_dump()
        == VerificationInput(
            files={
                pathlib.Path("libfile.py"): VerificationFile(
                    dependencies={pathlib.Path("libfile2.py")}
                ),
                pathlib.Path("libfile2.py"): VerificationFile(),
                pathlib.Path("test/test.py"): VerificationFile(),
            }
        ).model_dump()
    )


@pytest.mark.parametrize(
    ("obj", "results", "expected"),
    [
        (
            VerificationInput(
                files={
                    pathlib.Path("both/b1.c"): VerificationFile(),
                    pathlib.Path("both/b2.c"): VerificationFile(),
                    pathlib.Path("input/in1.c"): VerificationFile(),
                    pathlib.Path("input/in2.c"): VerificationFile(),
                }
            ),
            {
                pathlib.Path("both/b1.c"): FileResult(),
                pathlib.Path("both/b2.c"): FileResult(),
                pathlib.Path("result/res1.c"): FileResult(),
                pathlib.Path("result/res2.c"): FileResult(),
            },
            [
                (pathlib.Path("both/b1.c"), FileResult()),
                (pathlib.Path("both/b2.c"), FileResult()),
            ],
        ),
        (
            VerificationInput(
                files={
                    pathlib.Path("input/in1.c"): VerificationFile(),
                    pathlib.Path("input/in2.c"): VerificationFile(),
                }
            ),
            {
                pathlib.Path("result/res1.c"): FileResult(),
                pathlib.Path("result/res2.c"): FileResult(),
            },
            [],
        ),
    ],
)
def test_filterd_files(
    obj: VerificationInput,
    results: dict[ForcePosixPath, FileResult],
    expected: list[tuple[ForcePosixPath, FileResult]],
):
    assert list(obj.filterd_files(results)) == expected


def test_transitive_depends_on():
    obj = VerificationInput(
        files={
            pathlib.Path("fileA.c"): VerificationFile(
                dependencies={pathlib.Path("fileB.c"), pathlib.Path("other.c")}
            ),
            pathlib.Path("fileB.c"): VerificationFile(
                dependencies={pathlib.Path("fileA.c")}
            ),
            pathlib.Path("file1.c"): VerificationFile(),
            pathlib.Path("file2.c"): VerificationFile(
                dependencies={pathlib.Path("file1.c")}
            ),
            pathlib.Path("file3.c"): VerificationFile(
                dependencies={pathlib.Path("file1.c")}
            ),
            pathlib.Path("file4.c"): VerificationFile(
                dependencies={pathlib.Path("file2.c")}
            ),
        }
    )
    assert obj.transitive_depends_on == {
        pathlib.Path("fileA.c"): {pathlib.Path("fileA.c"), pathlib.Path("fileB.c")},
        pathlib.Path("fileB.c"): {pathlib.Path("fileA.c"), pathlib.Path("fileB.c")},
        pathlib.Path("file1.c"): {pathlib.Path("file1.c")},
        pathlib.Path("file2.c"): {pathlib.Path("file1.c"), pathlib.Path("file2.c")},
        pathlib.Path("file3.c"): {pathlib.Path("file1.c"), pathlib.Path("file3.c")},
        pathlib.Path("file4.c"): {
            pathlib.Path("file1.c"),
            pathlib.Path("file2.c"),
            pathlib.Path("file4.c"),
        },
    }



import logging
import os
import pathlib
import re
from datetime import datetime, timedelta, timezone

import pytest
from pytest_mock import MockerFixture

from competitive_verifier import app
from competitive_verifier.log import GitHubMessageParams
from competitive_verifier.models import (
    FileResult,
    ResultStatus,
    VerificationResult,
    VerifyCommandResult,
)
from competitive_verifier.verify import Verify
from competitive_verifier.verify.verifier import SplitState
from tests import LogComparer

test_get_split_state_params = [
    (None, None, None),
    (5, 0, SplitState(size=5, index=0)),
    (5, 1, SplitState(size=5, index=1)),
    (5, 2, SplitState(size=5, index=2)),
    (5, 3, SplitState(size=5, index=3)),
    (5, 4, SplitState(size=5, index=4)),
]


@pytest.mark.parametrize(
    ("size", "index", "expected"),
    test_get_split_state_params,
    ids=str,
)
def test_get_split_state(
    size: int | None,
    index: int | None,
    expected: SplitState | None,
):
    v = Verify(
        subcommand="verify",
        verify_files_json=pathlib.Path("verify.json"),
        split=size,
        split_index=index,
    )
    assert v.split_state == expected


test_get_split_state_error_params = {
    "No split index": (
        ["--verify-json", "verify.json", "--split", "2"],
        "--split argument requires --split-index argument.",
    ),
    "No split": (
        ["--verify-json", "verify.json", "--split-index", "2"],
        "--split-index argument requires --split argument.",
    ),
    "split index": (
        ["--verify-json", "verify.json", "--split-index", "5", "--split", "5"],
        "--split-index must be greater than 0 and less than --split.",
    ),
    "split index negative": (
        ["--verify-json", "verify.json", "--split-index", "-1", "--split", "5"],
        "--split-index must be greater than 0 and less than --split.",
    ),
    "split zero": (
        ["--verify-json", "verify.json", "--split-index", "1", "--split", "0"],
        "--split must be greater than 0.",
    ),
}


@pytest.mark.parametrize(
    ("args", "message"),
    test_get_split_state_error_params.values(),
    ids=test_get_split_state_error_params.keys(),
)
def test_get_split_state_error(args: list[str], message: str):
    parsed = app.ArgumentParser().parse(["verify", *args])

    assert isinstance(parsed, app.Verify)

    with pytest.raises(ValueError, match=rf"^{re.escape(message)}$"):
        _ = parsed.split_state


def test_invalid_prev_result(
    testtemp: pathlib.Path,
    caplog: pytest.LogCaptureFixture,
):
    (testtemp / "prev.json").write_bytes(b'[1,2,3,"invalid"]')
    parsed = app.ArgumentParser().parse(
        [
            "verify",
            "--verify-json",
            "verify.json",
            "--prev-result",
            str(testtemp / "prev.json"),
        ]
    )
    assert isinstance(parsed, app.Verify)

    assert parsed.read_prev_result() is None

    assert caplog.records == [
        LogComparer(
            f"Failed to parse prev_result: {testtemp / 'prev.json'}",
            logging.WARNING,
            github=GitHubMessageParams(file=testtemp / "prev.json"),
        )
    ]


def test_prev_result(testtemp: pathlib.Path):
    parsed = app.ArgumentParser().parse(["verify", "--verify-json", "verify.json"])
    assert isinstance(parsed, app.Verify)
    assert parsed.read_prev_result() is None

    (testtemp / "prev.json").write_text(
        '{"total_seconds":1.25,"files":{'
        '"' + (testtemp / "file1.txt").as_posix() + '"'
        ':{"verifications":[{"verification_name":"v1","status":"success","elapsed":1.2,"last_execution_time":"2025-09-21T08:09:10.001131-02:00"}],"newest":false},'
        '"file2.txt":{"verifications":[{"verification_name":"v2","status":"success","elapsed":1.3,"last_execution_time":"2025-09-21T08:09:11.004253-09:00"}],"newest":true}}}'
    )
    parsed = app.ArgumentParser().parse(
        [
            "verify",
            "--verify-json",
            "verify.json",
            "--prev-result",
            str(testtemp / "prev.json"),
        ]
    )
    assert isinstance(parsed, app.Verify)

    assert parsed.read_prev_result() == VerifyCommandResult(
        total_seconds=1.25,
        files={
            pathlib.Path("file1.txt"): FileResult(
                newest=False,
                verifications=[
                    VerificationResult(
                        verification_name="v1",
                        status=ResultStatus.SUCCESS,
                        last_execution_time=datetime(
                            2025,
                            9,
                            21,
                            8,
                            9,
                            10,
                            1131,
                            tzinfo=timezone(timedelta(hours=-2)),
                        ),
                        elapsed=1.2,
                    ),
                ],
            ),
            pathlib.Path("file2.txt"): FileResult(
                newest=True,
                verifications=[
                    VerificationResult(
                        verification_name="v2",
                        status=ResultStatus.SUCCESS,
                        last_execution_time=datetime(
                            2025,
                            9,
                            21,
                            8,
                            9,
                            11,
                            4253,
                            tzinfo=timezone(timedelta(hours=-9)),
                        ),
                        elapsed=1.3,
                    ),
                ],
            ),
        },
    )


@pytest.mark.allow_mkdir
@pytest.mark.parametrize("output", [True, False])
def test_write_result_output(
    output: bool,
    monkeypatch: pytest.MonkeyPatch,
    mocker: MockerFixture,
    testtemp: pathlib.Path,
    capsys: pytest.CaptureFixture[str],
):
    monkeypatch.chdir(testtemp)
    mocker.patch.dict(
        os.environ, {"GITHUB_STEP_SUMMARY": str(testtemp / "summary.md")}, clear=True
    )
    parsed = app.ArgumentParser().parse(
        [
            "verify",
            "--write-summary",
            "--verify-json",
            "verify.json",
            *(["--output", "result.json"] if output else []),
        ]
    )
    assert isinstance(parsed, app.Verify)
    parsed.write_result(
        VerifyCommandResult(
            total_seconds=5e-5,
            files={
                pathlib.Path("lib.c"): FileResult(
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SKIPPED,
                            elapsed=5e-6,
                            last_execution_time=datetime(
                                2021, 2, 28, 9, 17, tzinfo=timezone(timedelta(hours=9))
                            ),
                        )
                    ],
                )
            },
        )
    )
    expected = """{"total_seconds":0.00005,"files":{"lib.c":{"verifications":[{"status":"skipped","elapsed":5e-6,"last_execution_time":"2021-02-28T09:17:00+09:00"}],"newest":true}}}"""
    out, err = capsys.readouterr()
    assert out.strip() == expected
    assert err == ""

    if output:
        assert (testtemp / "result.json").read_text("utf-8") == expected
    else:
        assert not (testtemp / "result.json").exists()

    assert (
        (testtemp / "summary.md").read_text("utf-8")
        == """# ⚠ Verification result

- ✔&nbsp;&nbsp;All test case results are `success`
- ❌&nbsp;&nbsp;Test case results containts `failure`
- ⚠&nbsp;&nbsp;Test case results containts `skipped`


## Results
|📝&nbsp;&nbsp;File|✔<br>Passed|❌<br>Failed|⚠<br>Skipped|∑<br>Total|⏳<br>Elapsed|🦥<br>Slowest|🐘<br>Heaviest|
|:---|:---:|:---:|:---:|:---:|:---:|:---:|:---:|
|_**Sum**_|-|-|1|1|0ms|-|-|
|||||||||
|⚠&nbsp;&nbsp;lib.c|-|-|1|1|0ms|-|-|
"""
    )



from abc import ABC, abstractmethod
from typing import Annotated, Literal, Protocol

from pydantic import BaseModel, Field

from .path import ForcePosixPath
from .problem import TestCaseProvider
from .result import VerificationResult
from .result_status import ResultStatus
from .shell import ShellCommand, ShellCommandLike


class VerifcationTimeoutError(Exception):
    pass


class VerificationParams(Protocol):
    default_tle: float | None
    default_mle: float | None


class BaseVerification(BaseModel, ABC):
    name: str | None = None

    @abstractmethod
    def run(
        self,
        params: VerificationParams | None = None,
        *,
        deadline: float = float("inf"),
    ) -> ResultStatus | VerificationResult: ...

    @abstractmethod
    def run_compile_command(
        self,
        params: VerificationParams | None = None,
    ) -> bool: ...

    @property
    def is_lightweight(self) -> bool:
        """The verification is lightweight."""
        return False


class ConstVerification(BaseVerification):
    type: Literal["const"] = "const"
    status: ResultStatus = Field(description="The pre-defined result.")
    """The pre-defined result.
    """

    @property
    def is_lightweight(self) -> bool:
        return True

    def run(
        self,
        params: VerificationParams | None = None,
        *,
        deadline: float = float("inf"),
    ) -> ResultStatus:
        return self.status

    def run_compile_command(
        self,
        params: VerificationParams | None = None,
    ) -> bool:
        return True


class CommandVerification(BaseVerification):
    type: Literal["command"] = "command"

    command: ShellCommandLike = Field(description="The shell command for verification.")
    """The shell command for verification.
    """
    compile: ShellCommandLike | None = Field(
        default=None,
        description="The shell command for compile.",
    )
    """The shell command for compile.
    """

    tempdir: ForcePosixPath | None = Field(
        default=None,
        description="The temporary directory for running verification.",
    )
    """The temporary directory for running verification.
    """

    def run(
        self,
        params: VerificationParams | None = None,
        *,
        deadline: float = float("inf"),
    ) -> ResultStatus:
        if self.tempdir:
            self.tempdir.mkdir(parents=True, exist_ok=True)
        c = ShellCommand.parse_command_like(self.command)
        if c.exec_command(text=True).returncode == 0:
            return ResultStatus.SUCCESS
        return ResultStatus.FAILURE

    def run_compile_command(
        self,
        params: VerificationParams | None = None,
    ) -> bool:
        if self.compile:
            if self.tempdir:
                self.tempdir.mkdir(parents=True, exist_ok=True)
            c = ShellCommand.parse_command_like(self.compile)
            return c.exec_command(text=True).returncode == 0
        return True


class BaseProblemVerification(BaseVerification, ABC):
    command: ShellCommandLike = Field(description="The shell command for verification.")
    """The shell command for verification.
    """
    compile: ShellCommandLike | None = Field(
        default=None,
        description="The shell command for compile.",
    )
    """The shell command for compile.
    """

    error: float | None = Field(
        default=None,
        examples=[1e-9],
        description="The absolute or relative error to be considered as correct.",
    )
    """The absolute or relative error to be considered as correct.
    """
    tle: float | None = Field(
        default=None,
        examples=[10],
        description="The TLE time in seconds.",
    )
    """The TLE time in seconds.
    """
    mle: float | None = Field(
        default=None,
        examples=[64],
        description="The MLE memory size in megabytes.",
    )
    """The MLE memory size in megabytes.
    """

    @abstractmethod
    def _problem(self) -> TestCaseProvider | None: ...

    def run(
        self,
        params: VerificationParams | None = None,
        *,
        deadline: float = float("inf"),
    ) -> VerificationResult | ResultStatus:
        from competitive_verifier import oj  # noqa: PLC0415

        if not params:
            raise ValueError("ProblemVerification.run requires VerificationParams")

        problem = self._problem()
        if not problem:
            return ResultStatus.FAILURE

        c = ShellCommand.parse_command_like(self.command)
        result = oj.test(
            problem=problem,
            command=c.command,
            env=c.env,
            tle=self.tle or params.default_tle,
            error=self.error,
            mle=self.mle or params.default_mle,
            deadline=deadline,
        )
        result.verification_name = self.name
        return result

    def run_compile_command(
        self,
        params: VerificationParams | None = None,
    ) -> bool:
        if self.compile:
            c = ShellCommand.parse_command_like(self.compile)
            return c.exec_command(text=True).returncode == 0
        return True


class ProblemVerification(BaseProblemVerification):
    type: Literal["problem"] = "problem"

    problem: str = Field(
        description="The URL of problem.",
    )
    """
    problem: URL of problem
    """

    def _problem(self) -> TestCaseProvider | None:
        # circular dependency
        from competitive_verifier.oj import problem_from_url  # noqa: PLC0415

        return problem_from_url(self.problem)


class LocalProblemVerification(BaseProblemVerification):
    type: Literal["local"] = "local"

    input: ForcePosixPath = Field(
        description="The file path of testcases.",
    )
    """
    input: file path of testcases
    """

    tempdir: ForcePosixPath | None = Field(
        default=None,
        description="The temporary directory for running verification.",
    )
    """The temporary directory for running verification.
    """

    def _problem(self) -> TestCaseProvider | None:
        # circular dependency
        from competitive_verifier.oj import LocalProblem  # noqa: PLC0415

        return LocalProblem(self.input)

    def run_compile_command(self, params: VerificationParams | None = None) -> bool:
        if self.tempdir is not None:
            self.tempdir.mkdir(parents=True, exist_ok=True)

        return super().run_compile_command(params)


Verification = Annotated[
    ConstVerification
    | CommandVerification
    | ProblemVerification
    | LocalProblemVerification,
    Field(discriminator="type"),
]



import logging

from competitive_verifier.log import GitHubMessageParams
from tests import LogComparer


def test_comparer():
    assert LogComparer(message="123") == LogComparer(message="123")
    assert LogComparer(message="123") != LogComparer(message="124")

    assert LogComparer(message="123") == logging.makeLogRecord(
        {"msg": "%d", "args": (123,)}
    )
    assert LogComparer(message="123") != logging.makeLogRecord(
        {"msg": "%d", "args": (124,)}
    )

    assert LogComparer(message="123", level=1) == logging.makeLogRecord(
        {"msg": "%d", "args": (123,), "levelno": 1}
    )
    assert LogComparer(message="123", level=1) != logging.makeLogRecord(
        {"msg": "%d", "args": (123,), "levelno": 2}
    )

    assert LogComparer(message="123", name="1") == logging.makeLogRecord(
        {"msg": "%d", "args": (123,), "name": "1"}
    )
    assert LogComparer(message="123", name="1") != logging.makeLogRecord(
        {"msg": "%d", "args": (123,), "name": "2"}
    )

    assert LogComparer(
        message="123", github=GitHubMessageParams(title="air")
    ) == logging.makeLogRecord(
        {"msg": "%d", "args": (123,), "github": GitHubMessageParams(title="air")}
    )
    assert LogComparer(
        message="123", github=GitHubMessageParams(title="air2")
    ) != logging.makeLogRecord(
        {"msg": "%d", "args": (123,), "github": GitHubMessageParams(title="air")}
    )
    assert LogComparer(message="123") != logging.makeLogRecord(
        {"msg": "%d", "args": (123,), "github": GitHubMessageParams(title="air")}
    )



from typing import Literal

from pydantic import Field

from competitive_verifier.models import ShellCommand, ShellCommandLike

from .base import OjVerifyUserDefinedConfig
from .user_defined import UserDefinedLanguage


class OjVerifyHaskellConfig(OjVerifyUserDefinedConfig):
    execute: ShellCommandLike = Field(
        default_factory=lambda: ShellCommand(
            command=["runghc", "{basedir}/{path}"],
        ),
    )


class HaskellLanguage(UserDefinedLanguage):
    extension: Literal["hs"] = "hs"  # pyright: ignore[reportIncompatibleVariableOverride]
    config: OjVerifyHaskellConfig = Field(default_factory=OjVerifyHaskellConfig)  # pyright: ignore[reportIncompatibleVariableOverride]



import importlib.metadata
from argparse import ArgumentParser as BaseParser
from logging import getLogger
from typing import Annotated, Any, Literal, get_args

from pydantic import Field, TypeAdapter

from competitive_verifier.arg import BaseArguments
from competitive_verifier.documents import Docs
from competitive_verifier.download import Download
from competitive_verifier.inout import Check, MergeInput, MergeResult
from competitive_verifier.migrate import Migration
from competitive_verifier.oj import OjResolve
from competitive_verifier.verify import Verify

logger = getLogger(__name__)


class HelpError(Exception):
    pass


class NoSubcommand(BaseArguments):
    subcommand: Literal[None] = None  # noqa: PYI061

    def run(self) -> bool:
        raise HelpError

    @classmethod
    def add_parser(cls, parser: BaseParser):
        super().add_parser(parser)
        parser.add_argument(
            "--version",
            action="version",
            version=importlib.metadata.version("competitive-verifier"),
            help="print the competitive-verifier version number",
        )


Arguments = Annotated[
    NoSubcommand
    | Verify
    | Docs
    | Download
    | MergeInput
    | MergeResult
    | Check
    | OjResolve
    | Migration,
    Field(discriminator="subcommand"),
]
ARG_TYPES: tuple[type[BaseArguments], ...] = get_args(Arguments.__origin__)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]


class ArgumentParser(BaseParser):
    def __init__(self, **kwargs: dict[str, Any]) -> None:
        super().__init__(**kwargs)  # pyright: ignore[reportArgumentType]
        subparsers = self.add_subparsers(dest="subcommand", parser_class=BaseParser)

        for s in ARG_TYPES:
            s.add_parser(
                self
                if (sub := s.get_subcommand_info()) is None
                else subparsers.add_parser(**sub)
            )

    def parse(self, args: list[str] | None = None) -> Arguments:
        type_adapter = TypeAdapter[Arguments](Arguments)
        return type_adapter.validate_python(self.parse_args(args).__dict__)


def main(args: list[str] | None = None) -> int | None:
    parser = ArgumentParser()

    try:
        return not parser.parse(args).run()
    except HelpError:
        parser.print_help()
        return 2


if __name__ == "__main__":  # pragma: no cover
    main()



import json
import pathlib
from typing import Any

import pytest
from pydantic import ValidationError

from competitive_verifier.app import MergeInput
from competitive_verifier.models import (
    ConstVerification,
    ResultStatus,
    VerificationFile,
    VerificationInput,
)

test_merge_input_params: list[tuple[dict[str, VerificationInput], dict[str, Any]]] = [
    (
        {
            "file1.json": VerificationInput(
                files={
                    pathlib.Path("file1.txt"): VerificationFile(
                        dependencies={pathlib.Path("file2.txt")},
                        verification=ConstVerification(status=ResultStatus.SUCCESS),
                    ),
                    pathlib.Path("file2.txt"): VerificationFile(
                        document_attributes={"TITLE": "file2", "JSON": "file1"},
                    ),
                }
            ),
        },
        {
            "files": {
                "file1.txt": {
                    "dependencies": ["file2.txt"],
                    "verification": {"type": "const", "status": "success"},
                    "document_attributes": {},
                    "additonal_sources": [],
                },
                "file2.txt": {
                    "dependencies": [],
                    "verification": [],
                    "document_attributes": {"TITLE": "file2", "JSON": "file1"},
                    "additonal_sources": [],
                },
            }
        },
    ),
    (
        {
            "file1.json": VerificationInput(
                files={
                    pathlib.Path("file1.txt"): VerificationFile(
                        dependencies={pathlib.Path("file2.txt")},
                        verification=ConstVerification(status=ResultStatus.SUCCESS),
                    ),
                    pathlib.Path("file2.txt"): VerificationFile(
                        document_attributes={"TITLE": "file2", "JSON": "file1"},
                    ),
                }
            ),
            "file2.json": VerificationInput(
                files={
                    pathlib.Path("{base}/fileA.txt"): VerificationFile(
                        dependencies={pathlib.Path("file2.txt")},
                        verification=ConstVerification(status=ResultStatus.SUCCESS),
                    ),
                    pathlib.Path("{base}/file2.txt"): VerificationFile(
                        document_attributes={"NAME": "FILE2", "JSON": "file2"},
                    ),
                }
            ),
        },
        {
            "files": {
                "file1.txt": {
                    "dependencies": ["file2.txt"],
                    "verification": {"type": "const", "status": "success"},
                    "document_attributes": {},
                    "additonal_sources": [],
                },
                "file2.txt": {
                    "dependencies": [],
                    "verification": [],
                    "document_attributes": {"NAME": "FILE2", "JSON": "file2"},
                    "additonal_sources": [],
                },
                "fileA.txt": {
                    "additonal_sources": [],
                    "dependencies": [
                        "file2.txt",
                    ],
                    "document_attributes": {},
                    "verification": {
                        "status": "success",
                        "type": "const",
                    },
                },
            }
        },
    ),
]


@pytest.mark.parametrize(("files", "expected"), test_merge_input_params)
def test_merge_input(
    files: dict[str, VerificationInput],
    expected: dict[str, Any],
    testtemp: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
):
    file_list: list[pathlib.Path] = []
    for k, v in files.items():
        p = testtemp / k
        file_list.append(p)
        j = v.model_dump_json()
        p.write_text(j.replace(r"{base}", testtemp.as_posix()), encoding="utf-8")

    assert MergeInput(verify_files_json=file_list).run()
    out, err = capsys.readouterr()
    assert err == ""
    assert json.loads(out) == expected


def test_merge_input_error(testtemp: pathlib.Path):
    (testtemp / "ok.json").write_text(
        r"""{
    "files":{}
}"""
    )
    (testtemp / "ng.json").write_text(
        r"""{
    "files":[]
}"""
    )
    with pytest.raises(
        ValidationError, match=r"validation error for VerificationInput"
    ):
        MergeInput(verify_files_json=[testtemp / "ok.json", testtemp / "ng.json"]).run()



import math
import pathlib
import re
from typing import Any

import pytest
from pytest_mock import MockerFixture

from competitive_verifier import app
from competitive_verifier.arg import COMPETITIVE_VERIFY_FILES_PATH


def test_app_help(capsys: pytest.CaptureFixture[str]):
    assert app.main([]) == 2

    out, err = capsys.readouterr()
    assert out == app.ArgumentParser().format_help()
    assert err == ""


@pytest.mark.parametrize(("return_value", "expected"), [(True, 0), (False, 1)])
def test_result(return_value: bool, expected: int, mocker: MockerFixture):
    mocker.patch.object(
        app.Download,
        "run",
        return_value=return_value,
    )

    assert app.main(["download"]) == expected


test_parse_args_params: list[
    tuple[dict[str, str] | None, list[str], dict[str, Any]]
] = [
    (None, [], {"subcommand": None}),
    (
        None,
        ["verify", "--verify-json", ".competitive-verifier/verify_files.json"],
        {
            "subcommand": "verify",
            "default_mle": None,
            "default_tle": None,
            "download": True,
            "ignore_error": True,
            "output": None,
            "prev_result": None,
            "split": None,
            "split_index": None,
            "timeout": math.inf,
            "verbose": False,
            "verify_files_json": pathlib.Path(
                ".competitive-verifier/verify_files.json"
            ),
            "write_summary": False,
        },
    ),
    (
        None,
        [
            "verify",
            "--verify-json",
            ".competitive-verifier/verify_files.json",
            "--timeout",
            "0",
        ],
        {
            "subcommand": "verify",
            "default_mle": None,
            "default_tle": None,
            "download": True,
            "ignore_error": True,
            "output": None,
            "prev_result": None,
            "split": None,
            "split_index": None,
            "timeout": math.inf,
            "verbose": False,
            "verify_files_json": pathlib.Path(
                ".competitive-verifier/verify_files.json"
            ),
            "write_summary": False,
        },
    ),
    (
        {COMPETITIVE_VERIFY_FILES_PATH: ".competitive-verifier/verify_files.json"},
        ["verify"],
        {
            "subcommand": "verify",
            "default_mle": None,
            "default_tle": None,
            "download": True,
            "ignore_error": True,
            "output": None,
            "prev_result": None,
            "split": None,
            "split_index": None,
            "timeout": math.inf,
            "verbose": False,
            "verify_files_json": pathlib.Path(
                ".competitive-verifier/verify_files.json"
            ),
            "write_summary": False,
        },
    ),
    (
        {COMPETITIVE_VERIFY_FILES_PATH: ".competitive-verifier/verify_files.json"},
        [
            "verify",
            "--mle",
            "1024.5",
            "--tle",
            "2.5",
            "--no-download",
            "--check-error",
            "--write-summary",
            "--verbose",
            "--split",
            "6",
            "--split-index",
            "6",
            "--timeout",
            "20.5",
            "--prev-result",
            ".competitive-verifier/prev.json",
            "--output",
            ".competitive-verifier/out.json",
        ],
        {
            "subcommand": "verify",
            "default_mle": 1024.5,
            "default_tle": 2.5,
            "download": False,
            "ignore_error": False,
            "output": pathlib.Path(".competitive-verifier/out.json"),
            "prev_result": pathlib.Path(".competitive-verifier/prev.json"),
            "split": 6,
            "split_index": 6,
            "timeout": 20.5,
            "verbose": True,
            "verify_files_json": pathlib.Path(
                ".competitive-verifier/verify_files.json"
            ),
            "write_summary": True,
        },
    ),
    (
        {COMPETITIVE_VERIFY_FILES_PATH: ".competitive-verifier/verify_files.json"},
        [
            "verify",
            "-o",
            ".competitive-verifier/out.json",
        ],
        {
            "subcommand": "verify",
            "default_mle": None,
            "default_tle": None,
            "download": True,
            "ignore_error": True,
            "output": pathlib.Path(".competitive-verifier/out.json"),
            "prev_result": None,
            "split": None,
            "split_index": None,
            "timeout": math.inf,
            "verbose": False,
            "verify_files_json": pathlib.Path(
                ".competitive-verifier/verify_files.json"
            ),
            "write_summary": False,
        },
    ),
    (
        None,
        [
            "docs",
            "--verify-json",
            ".competitive-verifier/verify_files.json",
            "results/result1.json",
        ],
        {
            "subcommand": "docs",
            "destination": pathlib.Path(".competitive-verifier/_jekyll"),
            "docs": None,
            "exclude": [],
            "include": [],
            "ignore_error": True,
            "result_json": [pathlib.Path("results/result1.json")],
            "verbose": False,
            "verify_files_json": pathlib.Path(
                ".competitive-verifier/verify_files.json"
            ),
            "write_summary": False,
        },
    ),
    (
        {COMPETITIVE_VERIFY_FILES_PATH: ".competitive-verifier/verify_files.json"},
        [
            "docs",
            "results/result1.json",
            "results/result2.json",
            "--check-error",
            "--include",
            "indir1",
            "indir2",
            "--exclude",
            "indir1/ext",
            "indir2/ext/*",
            "--docs",
            ".verify/olddocs",
            "--destination",
            ".verify/_output",
            "--verbose",
            "--write-summary",
        ],
        {
            "subcommand": "docs",
            "destination": pathlib.Path(".verify/_output"),
            "docs": pathlib.Path(".verify/olddocs"),
            "include": ["indir1", "indir2"],
            "exclude": ["indir1/ext", "indir2/ext/*"],
            "ignore_error": False,
            "result_json": [
                pathlib.Path("results/result1.json"),
                pathlib.Path("results/result2.json"),
            ],
            "verbose": True,
            "verify_files_json": pathlib.Path(
                ".competitive-verifier/verify_files.json"
            ),
            "write_summary": True,
        },
    ),
    (
        None,
        ["download"],
        {
            "subcommand": "download",
            "verbose": False,
            "urls": [],
            "verify_files_json": None,
        },
    ),
    (
        None,
        [
            "download",
            "https://example.com/ex1",
            "--verify-json",
            ".competitive-verifier/verify_files.json",
        ],
        {
            "subcommand": "download",
            "verbose": False,
            "urls": ["https://example.com/ex1"],
            "verify_files_json": pathlib.Path(
                ".competitive-verifier/verify_files.json"
            ),
        },
    ),
    (
        {COMPETITIVE_VERIFY_FILES_PATH: ".competitive-verifier/verify_files.json"},
        ["download", "https://example.com/ex1", "https://example.com/ex2", "--verbose"],
        {
            "subcommand": "download",
            "verbose": True,
            "urls": ["https://example.com/ex1", "https://example.com/ex2"],
            "verify_files_json": None,
        },
    ),
    (
        {COMPETITIVE_VERIFY_FILES_PATH: ""},
        ["download", "https://example.com/ex1", "https://example.com/ex2"],
        {
            "subcommand": "download",
            "verbose": False,
            "urls": ["https://example.com/ex1", "https://example.com/ex2"],
            "verify_files_json": None,
        },
    ),
    (
        {COMPETITIVE_VERIFY_FILES_PATH: ".competitive-verifier/verify_files.json"},
        ["merge-input", "input/verify1.json", "input/verify2.json"],
        {
            "subcommand": "merge-input",
            "verbose": False,
            "verify_files_json": [
                pathlib.Path("input/verify1.json"),
                pathlib.Path("input/verify2.json"),
            ],
        },
    ),
    (
        None,
        ["merge-result", "result/result1.json"],
        {
            "subcommand": "merge-result",
            "write_summary": False,
            "verbose": False,
            "result_json": [pathlib.Path("result/result1.json")],
        },
    ),
    (
        None,
        [
            "merge-result",
            "--verbose",
            "result/result1.json",
            "result/result2.json",
            "--write-summary",
        ],
        {
            "subcommand": "merge-result",
            "write_summary": True,
            "verbose": True,
            "result_json": [
                pathlib.Path("result/result1.json"),
                pathlib.Path("result/result2.json"),
            ],
        },
    ),
    (
        None,
        ["check", "result/result1.json"],
        {
            "subcommand": "check",
            "verbose": False,
            "result_json": [pathlib.Path("result/result1.json")],
        },
    ),
    (
        None,
        ["check", "--verbose", "result/result1.json", "result/result2.json"],
        {
            "subcommand": "check",
            "verbose": True,
            "result_json": [
                pathlib.Path("result/result1.json"),
                pathlib.Path("result/result2.json"),
            ],
        },
    ),
    (
        None,
        ["oj-resolve"],
        {
            "subcommand": "oj-resolve",
            "verbose": False,
            "bundle": True,
            "config": None,
            "exclude": [],
            "include": [],
        },
    ),
    (
        None,
        [
            "oj-resolve",
            "--config",
            "new-config.toml",
            "--verbose",
            "--include",
            "indir1",
            "indir2",
            "--exclude",
            "indir1/ext",
            "indir2/ext/*",
            "--no-bundle",
        ],
        {
            "subcommand": "oj-resolve",
            "verbose": True,
            "bundle": False,
            "config": pathlib.Path("new-config.toml"),
            "include": ["indir1", "indir2"],
            "exclude": ["indir1/ext", "indir2/ext/*"],
        },
    ),
    (
        None,
        ["migrate"],
        {
            "subcommand": "migrate",
            "dry_run": False,
            "verbose": False,
        },
    ),
    (
        None,
        ["migrate", "--verbose", "--dry-run"],
        {
            "subcommand": "migrate",
            "dry_run": True,
            "verbose": True,
        },
    ),
]


@pytest.mark.parametrize(
    ("mockenv", "args", "expected"),
    test_parse_args_params,
    ids=(f"{' '.join(t[1])}:{t[0]}" for t in test_parse_args_params),
    indirect=["mockenv"],
)
@pytest.mark.usefixtures("mockenv")
def test_parse_args(
    args: list[str],
    expected: dict[str, Any],
    mocker: MockerFixture,
):
    parsed = app.ArgumentParser().parse(args)
    assert parsed.model_dump() == expected

    if not isinstance(parsed, app.NoSubcommand):
        mock_run = mocker.patch.object(parsed, "_run", return_value=True)
        mock_configure_stderr_logging = mocker.patch(
            "competitive_verifier.arg.configure_stderr_logging"
        )

        parsed.run()
        mock_run.assert_called_once()
        mock_configure_stderr_logging.assert_called_once_with(
            10 if parsed.verbose else 20
        )


test_parse_args_error_params: list[tuple[dict[str, str] | None, list[str], str]] = [
    (None, ["verify"], "the following arguments are required: --verify-json"),
    (
        None,
        ["docs"],
        "the following arguments are required: --verify-json, result_json",
    ),
    (
        {COMPETITIVE_VERIFY_FILES_PATH: ".competitive-verifier/verify_files.json"},
        ["docs"],
        "the following arguments are required: result_json",
    ),
    (None, ["merge-input"], "the following arguments are required: verify_files_json"),
    (None, ["merge-result"], "the following arguments are required: result_json"),
    (None, ["check"], "the following arguments are required: result_json"),
]


class ParseError(Exception):
    pass


@pytest.mark.parametrize(
    ("mockenv", "args", "expected_message"),
    test_parse_args_error_params,
    ids=[
        f"{' '.join(t[1])}:{' '.join((t[0] or {}).keys())}"
        for t in test_parse_args_error_params
    ],
    indirect=["mockenv"],
)
@pytest.mark.usefixtures("mockenv")
def test_parse_args_error(
    args: list[str],
    expected_message: str,
    mocker: MockerFixture,
):
    def _error(message: str):
        assert message
        raise ParseError(message)

    mocker.patch("argparse.ArgumentParser.error", side_effect=_error)
    with pytest.raises(ParseError, match=f"^{re.escape(expected_message)}$"):
        app.ArgumentParser().parse(args)



import pathlib

import competitive_verifier_resources


def test_doc_usage():
    doc_usage = competitive_verifier_resources.doc_usage(
        markdown_dir_path=pathlib.Path(),
        repo_name="example/myrepo",
    )
    assert "jekyll serve" in doc_usage
    assert "{{" not in doc_usage


def test_jekyll_files():
    assert "Gemfile" in competitive_verifier_resources.jekyll_files()

    for content in competitive_verifier_resources.jekyll_files().values():
        assert len(content) > 0



import pathlib
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable
from dataclasses import dataclass
from typing import TypeAlias

from colorama import Fore, Style

_whitespace_table = str.maketrans(
    {
        " ": "_",
        "\t": "\\t",
        "\r": "\\r",
        "\n": "\\n",
    }
)


def _replace_whitespace(s: str) -> str:
    return s.translate(_whitespace_table)


class _PrettyToken(ABC):
    """A token for verification output."""

    value: str

    def __init__(self, value: str) -> None:
        self.value = value

    @classmethod
    @abstractmethod
    def style(cls) -> str: ...

    def render(self) -> str:
        return self.style() + self.value + Style.RESET_ALL


class _BodyToken(_PrettyToken):
    @classmethod
    def style(cls) -> str:
        return Style.BRIGHT


class _HintToken(_PrettyToken):
    @classmethod
    def style(cls) -> str:
        return Style.DIM


class _WhitespaceToken(_PrettyToken):
    def __init__(self, value: str) -> None:
        self.value = _replace_whitespace(value)

    @classmethod
    def style(cls) -> str:
        return Style.DIM


_NewlineToken: TypeAlias = _WhitespaceToken


def _tokenize_str(s: str) -> list[_PrettyToken]:
    tokens: list[_PrettyToken] = []
    l = 0
    while l < len(s):
        r = l + 1
        while r < len(s) and (s[l] in " \t") == (s[r] in " \t"):
            r += 1
        typ = _WhitespaceToken if s[l] in " \t" else _BodyToken
        tokens.append(typ(s[l:r]))
        l = r
    return tokens


def _tokenize_line(line: str) -> list[_PrettyToken]:
    body = line.rstrip()
    newline = line[len(body) :]
    tokens: list[_PrettyToken] = []

    # add the body of line
    if body:
        tokens += _tokenize_str(body)

    # add newlines
    if newline:
        if newline in ("\n", "\r\n"):
            tokens.append(_NewlineToken(newline))
        else:
            whitespace = newline.rstrip("\r\n")
            newline = newline[len(whitespace) :]
            tokens.append(_WhitespaceToken(whitespace))
            tokens.append(_HintToken("(trailing whitespace)"))
            tokens.append(_NewlineToken(newline))

    return tokens


def _merge_token(tokens: list[_PrettyToken]) -> Iterable[_PrettyToken]:
    if len(tokens) == 0:
        yield _HintToken("(empty)")
        return

    prev = tokens[0]
    for token in tokens[1:]:
        if type(prev) is type(token):
            prev = type(prev)(prev.value + token.value)
        else:
            yield prev
            prev = token
    yield prev

    if isinstance(prev, _BodyToken):
        yield _HintToken("(no trailing newline)")


@dataclass
class Printer:
    content: str | pathlib.Path
    limit: int = 60
    head: int = 20
    tail: int = 20

    def __post_init__(self):  # pragma: no cover
        if self.head + self.tail >= self.limit:
            raise ValueError

    def __str__(self) -> str:
        return self.render_file_content()

    def render_file_content(self) -> str:
        tokens = self._tokenize_file_content()
        return "".join(token.render() for token in tokens)

    def _tokenize_file_content(self) -> Iterable[_PrettyToken]:
        if not isinstance(self.content, str):
            self.content = self.content.read_text()

        # Choose the shortest one from the three candidates.
        return _merge_token(list(self._token(self.content)))

    def _token(self, text: str) -> Generator[_PrettyToken, None, None]:
        if len(text) < self.limit:
            # short
            for line in text.splitlines(keepends=True):
                yield from _tokenize_line(line)
        else:
            # long
            left = text[: self.head]
            right = text[-self.tail :]

            for line in left.splitlines(keepends=True):
                yield from _tokenize_line(line)

            yield _HintToken(f"... ({len(text) - len(right) - len(left)} chars) ...")

            for line in right.splitlines(keepends=True):
                yield from _tokenize_line(line)


def green(s: str) -> str:
    return Fore.GREEN + s + Fore.RESET


def red(s: str) -> str:
    return Fore.RED + s + Fore.RESET



import math
import os
import pathlib
import platform
import shlex
import shutil
import subprocess
import sys
import tempfile
import time
from collections import Counter
from dataclasses import dataclass
from logging import getLogger
from typing import BinaryIO

from competitive_verifier.log import GitHubMessageParams
from competitive_verifier.models import (
    JudgeStatus,
    ResultStatus,
    TestCaseProvider,
    TestcaseResult,
    VerifcationTimeoutError,
    VerificationResult,
)

from . import gnu
from .format import Printer, green, red

logger = getLogger(__name__)


class CaseExecutionError(Exception):
    pass


@dataclass
class OjExecInfo:
    answer: str | None
    """The standard output of the executed command"""
    elapsed: float
    """The elapsed time of the executed command in seconds"""
    memory: float | None
    """The maximum memory usage of the executed command in megabytes"""
    returncode: int | None
    """The returncode of the executed command"""


def measure_command(
    command: list[str] | str,
    *,
    env: dict[str, str] | None = None,
    stdin: BinaryIO | int | None = None,
    timeout: float | None = None,
    gnu_time: bool = False,
) -> OjExecInfo:
    if isinstance(command, str):
        command = shlex.split(command)

    if len(command) == 0:
        raise CaseExecutionError

    with gnu.GnuTimeWrapper(enabled=gnu_time) as gw:
        if shutil.which(command[0]) is None:
            raise CaseExecutionError

        command = gw.get_command(command)
        begin = time.perf_counter()

        # We need kill processes called from the "time" command using process groups. Without this, orphans spawn. see https://github.com/kmyk/online-judge-tools/issues/640
        start_new_session = gw.gnu_time is not None

        try:
            if env is not None:
                env = os.environ | env
            proc = subprocess.run(
                command,
                env=env,
                timeout=timeout,
                stdin=stdin,
                stdout=subprocess.PIPE,
                stderr=sys.stderr,
                encoding="utf-8",
                start_new_session=start_new_session,
                check=False,
            )
            answer = proc.stdout
            returncode = proc.returncode
        except subprocess.TimeoutExpired:
            answer = None
            returncode = None
        except Exception as e:
            logger.exception(
                "'%s' is not executable.",
                command,
                extra={"github": GitHubMessageParams()},
            )
            raise CaseExecutionError from e

        end = time.perf_counter()
        return OjExecInfo(
            answer=answer,
            elapsed=end - begin,
            memory=gw.get_memory(),
            returncode=returncode,
        )


@dataclass
class OjTestArguments:
    """Parameters for oj-test command.

    Port of onlinejudge_command.subcommand.test.add_subparser.
    """

    command: str | list[str]
    problem: TestCaseProvider
    tle: float | None
    mle: float | None
    error: float | None
    env: dict[str, str] | None = None
    deadline: float = float("inf")


@dataclass
class OjTestcaseResult:
    name: str
    """A name of the test case."""
    input: pathlib.Path
    """A input of the test case."""
    answer: str
    """A output of the test case."""
    expected: pathlib.Path
    """A expected output of the test case."""

    status: JudgeStatus
    elapsed: float
    exitcode: int | None

    memory: float | None = None

    def __post_init__(self):
        if not isinstance(self.exitcode, int):
            self.exitcode = None

    def __str__(self) -> str:
        p = [
            f"{self.name}: {green('AC')}"
            if self.status == JudgeStatus.AC
            else f"{self.name}: {red(self.status.name)}",
            f"time: {self.elapsed:f} sec",
            f"memory: {self.memory:f} MB" if self.memory is not None else None,
            f"return code: {self.exitcode}" if self.exitcode else None,
        ]

        return ", ".join(filter(None, p))

    def log(self):
        match self.status:
            case JudgeStatus.AC:
                pass
            case JudgeStatus.RE | JudgeStatus.TLE:
                self._log_input()
                self._log_expected()
            case _:
                self._log_input()
                self._log_answer()
                self._log_expected()
        logger.info(self)

    def _log_input(self) -> None:
        logger.info("%s:input: %s", self.name, Printer(self.input))

    def _log_expected(self) -> None:
        logger.info("%s:expected: %s", self.name, Printer(self.expected))

    def _log_answer(self) -> None:
        logger.info("%s:answer: %s", self.name, Printer(self.answer))


@dataclass
class OjTestResult:
    is_success: bool

    elapsed: float

    slowest: float
    """max time [seconds]
    """

    heaviest: float
    """max memory [MB]
    """

    testcases: list[OjTestcaseResult]


def _try_parse_float(value: str) -> float | None:
    try:
        return float(value)
    except ValueError:
        return None


def _equal_or_closed_float(actual: str, expected: str, *, error: float) -> bool:
    if actual == expected:
        return True

    x = _try_parse_float(actual)
    y = _try_parse_float(expected)

    return (
        x is not None
        and y is not None
        and math.isclose(x, y, rel_tol=error, abs_tol=error)
    )


def compare_answer(actual: str, expected: str, *, error: float | None) -> bool:
    """Compare two byte strings.

    Args:
        actual (bytes): Actual output
        expected (bytes): Expected output
        error (float | None): Margin of error
    Returns:
        bool: True if they are considered equal
    """
    actual = actual.replace("\r\n", "\n")
    expected = expected.replace("\r\n", "\n")

    # match
    if actual == expected:
        return True

    try:
        if error is None:
            actual_words = actual.split()
            expected_words = expected.split()
            if all(x == y for x, y in zip(actual_words, expected_words, strict=True)):
                logger.warning("This was AC if spaces and newlines were ignored.")
            return False

        actual_lines = actual.rstrip("\n").split("\n")
        expected_lines = expected.rstrip("\n").split("\n")

        for actual_line, expected_line in zip(
            actual_lines, expected_lines, strict=True
        ):
            actual_words = actual_line.split()
            expected_words = expected_line.split()

            for x, y in zip(actual_words, expected_words, strict=True):
                if not _equal_or_closed_float(x, y, error=error):
                    return False
    except ValueError:
        return False

    return True


def special_judge(
    judge_command: str,
    output: str,
    *,
    input_path: pathlib.Path,
    expected_output_path: pathlib.Path | None,
) -> bool:
    with tempfile.TemporaryDirectory() as tempdir:
        actual_output_path = pathlib.Path(tempdir) / "actual.out"
        actual_output_path.write_text(output)

        command = [
            *shlex.split(judge_command),
            str(input_path.resolve()),
            str(actual_output_path.resolve()),
            str(
                expected_output_path.resolve()
                if expected_output_path is not None
                else ""
            ),
        ]

        logger.debug("$ %s", command)
        info = measure_command(command)
    logger.debug("judge's output: %s", Printer(info.answer or ""))
    return info.returncode == 0


def determine_status(
    *,
    exitcode: int | None,
    memory: float | None,
    mle: float | None,
    match_result: bool | None,
) -> JudgeStatus:
    if exitcode is None:
        return JudgeStatus.TLE
    if memory is not None and mle is not None and memory > mle:
        return JudgeStatus.MLE
    if exitcode != 0:
        return JudgeStatus.RE
    if match_result is not None and not match_result:
        return JudgeStatus.WA
    return JudgeStatus.AC


def single_case(
    test_name: str,
    test_input_path: pathlib.Path,
    test_output_path: pathlib.Path,
    *,
    args: OjTestArguments,
) -> OjTestcaseResult:
    try:
        logger.info("%s: start", test_name)

        # run the binary
        with test_input_path.open("rb") as infp:
            info = measure_command(
                args.command,
                env=args.env,
                stdin=infp,
                timeout=args.tle,
                gnu_time=True,
            )
            answer = info.answer or ""
            elapsed: float = info.elapsed
            memory: float | None = info.memory

        match_result = (
            special_judge(
                str(args.problem.checker),
                answer,
                input_path=test_input_path,
                expected_output_path=test_output_path,
            )
            if args.problem.checker
            else compare_answer(answer, test_output_path.read_text(), error=args.error)
        )

        status = determine_status(
            exitcode=info.returncode,
            memory=memory,
            mle=args.mle,
            match_result=match_result,
        )

        result = OjTestcaseResult(
            name=test_name,
            input=test_input_path,
            expected=test_output_path,
            answer=answer,
            status=status,
            exitcode=info.returncode,
            elapsed=elapsed,
            memory=memory,
        )
    except CaseExecutionError:
        logger.exception(
            "Failed to run: %s",
            args,
            extra={"github": GitHubMessageParams()},
        )
        return OjTestcaseResult(
            name=test_name,
            input=test_input_path,
            expected=test_output_path,
            answer="",
            status=JudgeStatus.RE,
            exitcode=255,
            elapsed=0,
            memory=None,
        )
    else:
        result.log()
        return result


def gnu_time_message(args: OjTestArguments):
    """Check wheather GNU time is available.

    Show messages if GNU time is not available.
    """
    if gnu.time_command() is None:
        if platform.system() == "Darwin":
            logger.info(
                "[HINT]: You can install GNU time with: $ brew install gnu-time",
                extra={"github": GitHubMessageParams()},
            )
        if args.mle is not None:
            logger.warning(
                "--mle is used but GNU time does not exist",
                extra={"github": GitHubMessageParams()},
            )


class _StatusCounter(Counter[JudgeStatus]):
    def __str__(self) -> str:
        return ", ".join(
            f"{cnt} {name}"
            for name, cnt in ((st.name, self.get(st)) for st in JudgeStatus)
            if cnt
        )


def summarize(history: list[OjTestcaseResult]):
    elapsed: float = 0.0
    slowest: float = -1.0
    slowest_name: str | None = None
    heaviest: float = -1.0
    heaviest_name: str | None = None
    counter = _StatusCounter()
    for result in history:
        counter[result.status] += 1
        elapsed += result.elapsed
        if slowest < result.elapsed:
            slowest = result.elapsed
            slowest_name = result.name
        if result.memory is not None and heaviest < result.memory:
            heaviest = result.memory
            heaviest_name = result.name

    # print the summary
    if slowest_name is not None:
        logger.info("slowest: %f sec  (for %s)", slowest, slowest_name)
    if heaviest_name is not None:
        logger.info("max memory: %f MB  (for %s)", heaviest, heaviest_name)

    length = len(history)
    is_success = counter[JudgeStatus.AC] == length
    if is_success:
        logger.info("%s %d cases", green("SUCCESS"), length)
    else:
        logger.info("%s %s / %d cases", red("FAILURE"), counter, length)

    # return the result
    return OjTestResult(
        is_success=is_success,
        slowest=slowest,
        elapsed=elapsed,
        heaviest=heaviest,
        testcases=history,
    )


def _run(args: OjTestArguments) -> OjTestResult:
    gnu_time_message(args)

    if args.error is not None and args.error > 1:
        logger.warning(
            "the tolerance is too large: relative = %s",
            args.error,
            extra={"github": GitHubMessageParams()},
        )

    tests = list(args.problem.iter_system_cases())

    # run tests
    history: list[OjTestcaseResult] = []
    for t in tests:
        if time.perf_counter() > args.deadline:
            raise VerifcationTimeoutError

        history.append(single_case(t.name, t.input_path, t.output_path, args=args))

    return summarize(history)


def main(
    *,
    problem: TestCaseProvider,
    command: str | list[str],
    env: dict[str, str] | None,
    tle: float | None,
    mle: float | None,
    error: float | None,
    deadline: float = float("inf"),
) -> VerificationResult:
    args = OjTestArguments(
        command=command,
        problem=problem,
        env=env,
        tle=tle,
        mle=mle,
        error=error,
        deadline=deadline,
    )
    result = _run(args)
    return VerificationResult(
        status=ResultStatus.SUCCESS if result.is_success else ResultStatus.FAILURE,
        elapsed=result.elapsed,
        slowest=result.slowest,
        heaviest=result.heaviest,
        testcases=[
            TestcaseResult(
                name=case.name,
                elapsed=case.elapsed,
                memory=case.memory,
                status=case.status,
            )
            for case in result.testcases
        ],
    )



import os
import pathlib
import re


def is_in_github_actions() -> bool:
    return os.getenv("GITHUB_ACTIONS", "").lower() == "true"


def get_ref_name() -> str | None:
    return os.getenv("GITHUB_REF_NAME")


def get_repository() -> str | None:
    return os.getenv("GITHUB_REPOSITORY")


def get_workflow_name() -> str | None:
    return os.getenv("GITHUB_WORKFLOW")


def get_workflow_ref() -> str | None:
    return os.getenv("GITHUB_WORKFLOW_REF")


def get_workflow_filename() -> str | None:
    ref = get_workflow_ref()
    if not ref:
        return None
    return re.sub(r".*/([^/]+\.yml)@.*$", r"\1", ref)


def _optional_path(strpath: str | None) -> pathlib.Path | None:
    return pathlib.Path(strpath) if strpath else None


def get_step_summary_path() -> pathlib.Path | None:
    return _optional_path(os.getenv("GITHUB_STEP_SUMMARY"))



from contextlib import nullcontext
from logging import getLogger

from competitive_verifier import log
from competitive_verifier.models import Problem

from .problem import NotLoggedInError, problem_from_url

logger = getLogger(__name__)


def _run(*, problem: Problem) -> bool:
    return bool(problem.download_system_cases())


def main(url: str, *, group_log: bool = False) -> bool:
    # prepare values
    problem = problem_from_url(url)
    if problem is None:
        logger.error(
            'The URL "%s" is not supported',
            url,
            extra={"github": log.GitHubMessageParams()},
        )
        return False

    with (
        log.group(f"download[Run]: {url}")
        if group_log
        else nullcontext(logger.info("download[Run]: %s", url))
    ):
        try:
            _run(problem=problem)
        except Exception as e:
            if isinstance(e, NotLoggedInError):
                logger.exception(
                    "Login is required to download the problem. %r",
                    url,
                    extra={"github": log.GitHubMessageParams()},
                )
            else:
                logger.exception(
                    "Failed to download. %r",
                    url,
                    extra={"github": log.GitHubMessageParams()},
                )
            return False
    return True



import pathlib
from contextlib import nullcontext
from subprocess import CompletedProcess
from unittest import mock

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.oj import gnu


@pytest.fixture(autouse=True)
def reset_time_command_cache():
    gnu.time_command.cache_clear()
    yield
    gnu.time_command.cache_clear()


@pytest.mark.parametrize(
    ("os_name", "expected"),
    [
        ("nt", ["time", "gtime"]),
        (
            "posix",
            ["time", "gtime", "/bin/time", "/usr/bin/time"],
        ),
    ],
)
def test_find_gnu_time_candidate(
    os_name: str,
    expected: list[str],
    mocker: MockerFixture,
):
    find_gnu_time = mocker.patch.object(gnu, "_find_gnu_time", return_value=None)
    with mock.patch("os.name", os_name):
        gnu.time_command()

    find_gnu_time.assert_called_once_with(expected)


@pytest.mark.parametrize(
    ("os_name", "expected"),
    [("nt", None), ("posix", "/bin/time")],
)
def test_find_gnu_time(os_name: str, expected: str, mocker: MockerFixture):
    def check(gnu_time: str) -> bool:
        return gnu_time == "/bin/time"

    mocker.patch.object(gnu, "check_gnu_time", side_effect=check)
    with mock.patch("os.name", os_name):
        assert gnu.time_command() == expected


test_check_gnu_time_params = [
    (
        CompletedProcess[str](
            "dummy",
            returncode=0,
            stdout="check_gnu_time",
        ),
        "1024 KB\n\n",
        True,
    ),
    (
        CompletedProcess[str](
            "dummy",
            returncode=0,
            stdout="error",
        ),
        "invalid",
        False,
    ),
    (
        CompletedProcess[str](
            "dummy",
            returncode=1,
            stdout="check_gnu_time",
        ),
        "1024 KB\n\n",
        False,
    ),
    (
        CompletedProcess[str](
            "dummy",
            returncode=0,
            stdout="check_gnu_time",
        ),
        "invalid",
        False,
    ),
    (
        FileNotFoundError,
        "invalid",
        False,
    ),
]


@pytest.mark.parametrize(
    ("command_result", "outdata", "expected"),
    test_check_gnu_time_params,
)
def test_check_gnu_time(
    command_result: CompletedProcess[str] | type[Exception],
    outdata: str,
    expected: bool,
    testtemp: pathlib.Path,
    caplog: pytest.LogCaptureFixture,
    mocker: MockerFixture,
):
    caplog.set_level(0)
    mocker.patch("tempfile.TemporaryDirectory", return_value=nullcontext(testtemp))

    if isinstance(command_result, type):
        mocker.patch("subprocess.run", side_effect=command_result)
    else:
        mocker.patch("subprocess.run", return_value=command_result)
    (testtemp / "out").write_text(outdata)
    assert gnu.check_gnu_time("dummy_time") == expected


@pytest.mark.parametrize("error", [NameError, AttributeError])
def test_check_gnu_time_error(error: type[Exception], mocker: MockerFixture):
    with mock.patch("os.name", "posix"):
        mocker.patch("tempfile.TemporaryDirectory", side_effect=error())
        with pytest.raises(error):
            gnu.time_command()


def test_gnu_wrapper_not_found(mocker: MockerFixture):
    mocker.patch.object(gnu, "time_command", return_value=None)
    with gnu.GnuTimeWrapper(enabled=True) as gw:
        cmd = gw.get_command(["foo", "bar"])
        assert cmd == ["foo", "bar"]
        assert gw.get_memory() is None


def test_gnu_wrapper(testtemp: pathlib.Path, mocker: MockerFixture):
    mocker.patch.object(gnu, "time_command", return_value="dummy_time")
    with gnu.GnuTimeWrapper(enabled=True) as gw:
        cmd = gw.get_command(["foo", "bar"])
        assert cmd == [
            "dummy_time",
            "-f",
            "%M",
            "-o",
            str(testtemp / "1" / "gnu_time_report.txt"),
            "--",
            "foo",
            "bar",
        ]
        outpath = cmd[4]
        assert gw.get_memory() is None
        pathlib.Path(outpath).write_text("2048")
        assert gw.get_memory() == 2.048
        pathlib.Path(outpath).write_text("Other output\n123456\n\n")
        assert gw.get_memory() == 123.456
        pathlib.Path(outpath).write_text("abc")
        assert gw.get_memory() is None

    with gnu.GnuTimeWrapper(enabled=False) as gw:
        cmd = gw.get_command(["foo", "bar"])
        assert cmd == ["foo", "bar"]
        assert gw.get_memory() is None



import datetime
import pathlib
import time
from abc import ABC, abstractmethod
from functools import cached_property
from logging import getLogger

from competitive_verifier import git, log
from competitive_verifier.download import download_files as run_download
from competitive_verifier.models import (
    FileResult,
    ResultStatus,
    VerifcationTimeoutError,
    Verification,
    VerificationFile,
    VerificationInput,
    VerificationResult,
    VerifyCommandResult,
)
from competitive_verifier.resource import try_ulimit_stack
from competitive_verifier.verify.split_state import SplitState

logger = getLogger(__name__)


def _now() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc).astimezone()


class InputContainer(ABC):
    verifications: VerificationInput
    verification_time: datetime.datetime
    prev_result: VerifyCommandResult | None
    split_state: SplitState | None

    def __init__(
        self,
        *,
        verifications: VerificationInput,
        verification_time: datetime.datetime,
        prev_result: VerifyCommandResult | None,
        split_state: SplitState | None,
    ) -> None:
        self.verifications = verifications
        self.verification_time = verification_time
        self.prev_result = prev_result
        self.split_state = split_state

    @abstractmethod
    def get_file_timestamp(self, path: pathlib.Path) -> datetime.datetime: ...

    def file_need_verification(
        self,
        path: pathlib.Path,
        file_result: FileResult,
    ) -> bool:
        if not path.exists():
            return False
        base_time = min(self.verification_time, self.get_file_timestamp(path))
        result = file_result.need_verification(base_time)
        if result:
            logger.info("%s needs verification. base_time: %s", path, base_time)
        else:
            logger.info("%s doesn't need verification. base_time: %s", path, base_time)
        return result

    @cached_property
    def verification_files(self) -> dict[pathlib.Path, VerificationFile]:
        """List of verification files."""
        return {
            p: f for p, f in self.verifications.files.items() if f.is_verification()
        }

    @cached_property
    def skippable_verification_files(self) -> dict[pathlib.Path, VerificationFile]:
        return {
            p: f
            for p, f in self.verification_files.items()
            if f.is_lightweight_verification()
        }

    @cached_property
    def remaining_verification_files(self) -> dict[pathlib.Path, VerificationFile]:
        """List of verification files that have not yet been verified."""
        verification_files = {
            p: f
            for p, f in self.verification_files.items()
            if p not in self.skippable_verification_files
        }

        if self.prev_result is None:
            return verification_files

        not_updated_files = {
            k
            for k, v in self.verifications.filterd_files(self.prev_result.files)
            if not self.file_need_verification(k, v)
        }
        return {
            p: f for p, f in verification_files.items() if p not in not_updated_files
        }

    @cached_property
    def current_verification_files(self) -> dict[pathlib.Path, VerificationFile]:
        """List of verification files that self should verify.

        if ``split_state`` is None the property is ``remaining_verification_files``;

        else ``split_state.split(remaining_verification_files)``.
        """
        if self.split_state is None:
            return self.remaining_verification_files

        lst = [(p, f) for p, f in self.remaining_verification_files.items()]
        lst.sort(key=lambda tup: tup[0])

        return dict(self.split_state.split(lst))


class BaseVerifier(InputContainer):
    timeout: float
    default_tle: float | None
    default_mle: float | None
    split_state: SplitState | None

    _result: VerifyCommandResult | None

    def __init__(
        self,
        verifications: VerificationInput,
        *,
        timeout: float,
        default_tle: float | None,
        default_mle: float | None,
        prev_result: VerifyCommandResult | None,
        split_state: SplitState | None,
        verification_time: datetime.datetime | None = None,
    ) -> None:
        super().__init__(
            verifications=verifications,
            verification_time=verification_time or _now(),
            prev_result=prev_result,
            split_state=split_state,
        )
        self._input = verifications
        self.timeout = timeout
        self.default_tle = default_tle
        self.default_mle = default_mle
        self._result = None

    @property
    def is_first(self) -> bool:
        if not self.split_state:
            return True
        return self.split_state.index == 0

    def _enumerate_verifications(
        self,
        p: pathlib.Path,
        f: VerificationFile,
        *,
        download: bool,
        deadline: float,
    ) -> list[VerificationResult]:
        logger.debug("%r", f)
        verifications = list[VerificationResult]()
        try:
            if time.perf_counter() > deadline:
                raise VerifcationTimeoutError  # noqa: TRY301
            if download:
                run_download(f, check=True, group_log=False)
        except VerifcationTimeoutError:
            verifications.append(
                self.create_command_result(ResultStatus.SKIPPED, time.perf_counter())
            )
            logger.warning("Skip[Timeout]: %s", p)
            return verifications
        except BaseException:
            verifications.append(
                self.create_command_result(ResultStatus.FAILURE, time.perf_counter())
            )
            logger.exception(
                "Failed to download: %s",
                f.verification,
                extra={"github": log.GitHubMessageParams()},
            )
            return verifications

        for ve in f.verification_list:
            logger.debug("command=%r", ve)
            prev_time = time.perf_counter()
            try:
                if prev_time > deadline:
                    raise VerifcationTimeoutError  # noqa: TRY301

                rs, error_message = self.run_verification(ve, deadline=deadline)
                if error_message:
                    logger.error(
                        "%s: %s, verification=%s",
                        error_message,
                        p,
                        ve.model_dump_json(exclude_unset=True),
                        extra={"github": log.GitHubMessageParams(file=p)},
                    )
                verifications.append(
                    self.create_command_result(rs, prev_time, name=ve.name)
                )
            except VerifcationTimeoutError:
                logger.warning("Skip[Timeout]: %s, %r", p, ve)
                verifications.append(
                    self.create_command_result(
                        ResultStatus.SKIPPED,
                        prev_time,
                        name=ve.name,
                    )
                )
            except BaseException:
                logger.exception(
                    "Failed to verify: %s, %r",
                    p,
                    ve,
                    extra={"github": log.GitHubMessageParams()},
                )
                verifications.append(
                    self.create_command_result(
                        ResultStatus.FAILURE,
                        prev_time,
                        name=ve.name,
                    )
                )
        return verifications

    def verify(self, *, download: bool = True) -> VerifyCommandResult:
        start_time = time.perf_counter()
        deadline = start_time + self.timeout

        with log.group("current_verification_files"):
            current_verification_files = self.current_verification_files
            logger.info(
                "current_verification_files: %s",
                " ".join(p.as_posix() for p in current_verification_files),
            )
        try_ulimit_stack()

        file_results: dict[pathlib.Path, FileResult] = (
            {
                k: v.model_copy(update={"newest": False})
                for k, v in self.verifications.filterd_files(self.prev_result.files)
                if k.exists()
            }
            if self.prev_result
            else {}
        )

        for p, f in current_verification_files.items():
            with log.group(f"Verify: {p.as_posix()}"):
                file_results[p] = FileResult(
                    verifications=self._enumerate_verifications(
                        p,
                        f,
                        download=download,
                        deadline=deadline,
                    )
                )

        sippable_file_results = self.skippable_results()
        self._result = VerifyCommandResult(
            total_seconds=time.perf_counter() - start_time,
            files=file_results | sippable_file_results,
        )
        return self._result

    def run_verification(
        self,
        verification: Verification,
        *,
        deadline: float = float("inf"),
    ) -> tuple[ResultStatus | VerificationResult, str | None]:
        """Run verification.

        Returns:
            tuple[ResultStatus, Optional[str]]: (Result, error_message)
        """
        if not verification.run_compile_command():
            return ResultStatus.FAILURE, "Failed to compile"

        if time.perf_counter() > deadline:
            raise VerifcationTimeoutError

        rs = verification.run(self, deadline=deadline)

        if rs.status != ResultStatus.SUCCESS:
            return rs, "Failed to test"
        return rs, None

    def skippable_results(self) -> dict[pathlib.Path, FileResult]:
        """Run skippable verification."""
        results = dict[pathlib.Path, FileResult]()
        if self.is_first:
            for p, f in self.skippable_verification_files.items():
                logger.info("Start skippable: %s", p)
                verifications = list[VerificationResult]()
                prev_time = time.perf_counter()

                for v in f.verification_list:
                    rs = self.run_verification(v)[0]
                    verifications.append(
                        self.create_command_result(rs, prev_time, name=v.name)
                    )
                results[p] = FileResult(
                    verifications=verifications,
                    newest=True,
                )
        return results

    def create_command_result(
        self,
        status_or_result: ResultStatus | VerificationResult,
        prev_time: float,
        *,
        name: str | None = None,
    ) -> VerificationResult:
        if isinstance(status_or_result, VerificationResult):
            return status_or_result

        elapsed = time.perf_counter() - prev_time
        return VerificationResult(
            verification_name=name,
            status=status_or_result,
            elapsed=elapsed,
            last_execution_time=self.verification_time,
        )


class Verifier(BaseVerifier):
    use_git_timestamp: bool

    def __init__(
        self,
        verifications: VerificationInput,
        *,
        timeout: float,
        default_tle: float | None,
        default_mle: float | None,
        prev_result: VerifyCommandResult | None,
        split_state: SplitState | None,
        verification_time: datetime.datetime | None = None,
        use_git_timestamp: bool,
    ) -> None:
        super().__init__(
            verifications=verifications,
            verification_time=verification_time or _now(),
            prev_result=prev_result,
            split_state=split_state,
            timeout=timeout,
            default_tle=default_tle,
            default_mle=default_mle,
        )
        self.use_git_timestamp = use_git_timestamp

    def get_file_timestamp(self, path: pathlib.Path) -> datetime.datetime:
        dependicies = self.verifications.transitive_depends_on[path]

        if self.use_git_timestamp:
            return git.get_commit_time(dependicies)

        timestamp = max(x.stat().st_mtime for x in dependicies)
        system_local_timezone = _now().tzinfo

        # microsecond=0 is required because it's erased in git commit
        return datetime.datetime.fromtimestamp(
            timestamp, tz=system_local_timezone
        ).replace(microsecond=0)



import logging
import pathlib
from collections.abc import Iterable

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.download.download import (
    Download,
    UrlOrVerificationFile,
    parse_urls,
)
from competitive_verifier.models import (
    ProblemVerification,
    VerificationFile,
    VerificationInput,
)
from tests import LogComparer


def get_problem_command(url: str) -> ProblemVerification:
    return ProblemVerification(command="true", problem=url)


_SomeUrlOrVerificationFile = UrlOrVerificationFile | Iterable[UrlOrVerificationFile]
test_parse_urls_params: list[tuple[_SomeUrlOrVerificationFile, set[str]]] = [
    (
        "http://example.com",
        {
            "http://example.com",
        },
    ),
    (
        VerificationFile(
            verification=[
                get_problem_command("http://example.com/alpha"),
                get_problem_command("http://example.com/beta"),
                get_problem_command("http://example.com/gamma"),
                get_problem_command("http://example.com/delta"),
            ]
        ),
        {
            "http://example.com/alpha",
            "http://example.com/beta",
            "http://example.com/gamma",
            "http://example.com/delta",
        },
    ),
    (
        [
            VerificationFile(
                verification=[
                    get_problem_command("http://example.com/alpha"),
                    get_problem_command("http://example.com/beta"),
                    get_problem_command("http://example.com/gamma"),
                    get_problem_command("http://example.com/delta"),
                ]
            ),
            VerificationFile(
                verification=[
                    get_problem_command("https://example.com/alpha"),
                    get_problem_command("https://example.com/beta"),
                    get_problem_command("https://example.com/gamma"),
                    get_problem_command("https://example.com/delta"),
                ]
            ),
        ],
        {
            "http://example.com/alpha",
            "http://example.com/beta",
            "http://example.com/gamma",
            "http://example.com/delta",
            "https://example.com/alpha",
            "https://example.com/beta",
            "https://example.com/gamma",
            "https://example.com/delta",
        },
    ),
    (
        [
            VerificationFile(
                verification=[
                    get_problem_command("http://example.com/alpha"),
                    get_problem_command("http://example.com/beta"),
                    get_problem_command("http://example.com/gamma"),
                    get_problem_command("http://example.com/delta"),
                    get_problem_command("https://example.com/alpha"),
                ]
            ),
            "https://example.com/alpha",
            "https://example.com/beta",
            "https://example.com/beta",
            "https://example.com/gamma",
            "https://example.com/delta",
        ],
        {
            "http://example.com/alpha",
            "http://example.com/beta",
            "http://example.com/gamma",
            "http://example.com/delta",
            "https://example.com/alpha",
            "https://example.com/beta",
            "https://example.com/gamma",
            "https://example.com/delta",
        },
    ),
]


@pytest.mark.parametrize(
    ("url_or_file", "expected"),
    test_parse_urls_params,
)
def test_parse_urls(
    url_or_file: _SomeUrlOrVerificationFile,
    expected: set[str],
):
    assert set(parse_urls(url_or_file)) == expected


test_download_run_params: list[
    tuple[list[str], VerificationInput | None, list[str | VerificationFile]]
] = [
    ([], None, []),
    (
        [
            "https://example.com/alpha",
            "https://example.com/beta",
            "https://example.com/beta",
            "https://example.com/gamma",
            "https://example.com/delta",
        ],
        None,
        [
            "https://example.com/alpha",
            "https://example.com/beta",
            "https://example.com/beta",
            "https://example.com/gamma",
            "https://example.com/delta",
        ],
    ),
    (
        [],
        VerificationInput(
            files={
                pathlib.Path("foo.py"): VerificationFile(
                    verification=[
                        get_problem_command("http://example.com/alpha"),
                        get_problem_command("http://example.com/beta"),
                        get_problem_command("http://example.com/gamma"),
                        get_problem_command("http://example.com/delta"),
                        get_problem_command("https://example.com/alpha"),
                    ]
                ),
            }
        ),
        [
            VerificationFile(
                verification=[
                    get_problem_command("http://example.com/alpha"),
                    get_problem_command("http://example.com/beta"),
                    get_problem_command("http://example.com/gamma"),
                    get_problem_command("http://example.com/delta"),
                    get_problem_command("https://example.com/alpha"),
                ]
            ),
        ],
    ),
    (
        [
            "https://example.com/kilo",
            "https://example.com/mega",
            "https://example.com/giga",
            "https://example.com/tera",
        ],
        VerificationInput(
            files={
                pathlib.Path("foo.py"): VerificationFile(
                    verification=[
                        get_problem_command("http://example.com/alpha"),
                        get_problem_command("http://example.com/beta"),
                        get_problem_command("http://example.com/gamma"),
                        get_problem_command("http://example.com/delta"),
                        get_problem_command("https://example.com/alpha"),
                    ]
                ),
            }
        ),
        [
            VerificationFile(
                verification=[
                    get_problem_command("http://example.com/alpha"),
                    get_problem_command("http://example.com/beta"),
                    get_problem_command("http://example.com/gamma"),
                    get_problem_command("http://example.com/delta"),
                    get_problem_command("https://example.com/alpha"),
                ]
            ),
            "https://example.com/kilo",
            "https://example.com/mega",
            "https://example.com/giga",
            "https://example.com/tera",
        ],
    ),
]


@pytest.mark.parametrize(
    ("urls", "verification_input", "expected"),
    test_download_run_params,
)
def test_download_run(
    urls: list[str],
    verification_input: VerificationInput | None,
    expected: list[str | VerificationFile],
    caplog: pytest.LogCaptureFixture,
    mocker: MockerFixture,
    testtemp: pathlib.Path,
):
    caplog.set_level(0)
    mock_download_files = mocker.patch(
        "competitive_verifier.download.download.download_files"
    )

    verify_files_json = None
    if verification_input:
        (verify_files_json := testtemp / "verify.json").write_text(
            verification_input.model_dump_json()
        )

    assert Download(urls=urls, verify_files_json=verify_files_json).run()

    mock_download_files.assert_called_once_with(expected, group_log=True)
    assert caplog.records == [
        LogComparer(
            f"arguments:{Download(urls=urls, verify_files_json=verify_files_json)}",
            logging.DEBUG,
        ),
        LogComparer(
            f"verify_files_json={verify_files_json}, urls={urls}",
            logging.INFO,
        ),
    ]



import pathlib
from typing import NamedTuple

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.util import (
    normalize_bytes_text,
    read_text_normalized,
    resolve_referenced_path,
)


class DirStructure(NamedTuple):
    name: str
    items: "list[str|DirStructure]"


@pytest.mark.allow_mkdir
def test_resolve_referenced_path(
    monkeypatch: pytest.MonkeyPatch,
    testtemp: pathlib.Path,
):
    def setup(cwd: pathlib.Path, d: DirStructure):
        dd = cwd / d.name
        dd.mkdir()
        for c in d.items:
            if isinstance(c, str):
                (dd / c).touch()
            else:
                setup(dd, c)

    setup(
        testtemp,
        DirStructure(
            "root",
            [
                DirStructure(
                    "dir1",
                    ["file1.txt", DirStructure("subdirA", ["a1.cb1.c"])],
                ),
                DirStructure(
                    "dir2",
                    ["file1.txt", DirStructure("subdirB", ["a1.cb1.c"])],
                ),
            ],
        ),
    )
    (testtemp / "outer").touch()
    monkeypatch.chdir(testtemp / "root")

    assert (
        resolve_referenced_path(str(testtemp / "outer"), basedir=testtemp / "root")
        is None
    )

    assert resolve_referenced_path(
        "./file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir1/file1.txt")

    assert resolve_referenced_path(
        "../dir2/file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir2/file1.txt")

    assert (
        resolve_referenced_path("./file2.txt", basedir=testtemp / "root/dir1") is None
    )

    assert (
        resolve_referenced_path("../dir2/file2.txt", basedir=testtemp / "root/dir1")
        is None
    )

    assert resolve_referenced_path(
        "//dir1/file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir1/file1.txt")

    assert resolve_referenced_path(
        "//dir2/file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir2/file1.txt")

    assert (
        resolve_referenced_path("//file2.txt", basedir=testtemp / "root/dir1") is None
    )

    assert resolve_referenced_path(
        "file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir1/file1.txt")

    assert resolve_referenced_path(
        str(testtemp / "root/dir1/file1.txt"), basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir1/file1.txt")


def test_normalize_bytes_text(mocker: MockerFixture):
    cp932_text = (
        b"\x89J\x83j\x83\x82\x83}\x83P\x83Y\n"
        b"\x95\x97\x83j\x83\x82\x83}\x83P\x83Y\n"
        b"\x90\xe1\x83j\x83\x82\x89\xc4\x83m\x8f\x8b\x83T\x83j\x83\x82\x83}\x83P\x83k\n"
        b"\x8f\xe4\x95v\x83i\x83J\x83\x89\x83_\x83\x92\x83\x82\x83`\n"
        b"\x97|\x83n\x83i\x83N\n\x8c\x88\x83V\x83e\xe1\xd1\x83\x89\x83Y\n"
        b"\x83C\x83c\x83\x82\x83V\x83d\x83J\x83j\x83\x8f\x83\x89\x83b\x83e\x83\x90\x83\x8b\n"
        b"\x88\xea\x93\xfa\x83j\x8c\xba\x95\xc4\x8el\x8d\x87\x83g\n"
        b"\x96\xa1\x91X\x83g\x8f\xad\x83V\x83m\x96\xec\x8d\xd8\x83\x92\x83^\x83x\n"
        b"\x83A\x83\x89\x83\x86\x83\x8b\x83R\x83g\x83\x92\n"
        b"\x83W\x83u\x83\x93\x83\x92\x83J\x83\x93\x83W\x83\x87\x83E\x83j\x93\xfc\x83\x8c\x83Y\x83j\n"
        b"\x83\x88\x83N\x83~\x83L\x83L\x83V\x83\x8f\x83J\x83\x8a\n"
        b"\x83\\\x83V\x83e\x83\x8f\x83X\x83\x8c\x83Y\n"
        b"\x96\xec\x8c\xb4\xc9\x8f\xbc\xc9\x97\xd1\xc9\x89A\xc9\n"
        b"\x8f\xac\xbb\xc5\x8a\x9e\xcc\xde\xb7\xc9\x8f\xac\x89\xae\xc6\x83\x90\xc3\n"
        b"\x93\x8c\xc6\x95a\x8bC\xc9\xba\xc4\xde\xd3\xb1\xda\xca\xde\n"
        b"\x8ds\xaf\xc3\x8a\xc5\x95a\xbc\xc3\xd4\xd8\n"
        b"\x90\xbc\xc6\xc2\xb6\xda\xc0\x95\xea\xb1\xda\xca\xde\n"
        b"\x8ds\xaf\xc3\xbf\xc9\x88\xee\xc9\x91\xa9\xa6\x95\x89\xcb\n"
        b"\x93\xec\xc6\x8e\x80\xc6\xbb\xb3\xc5\x90l\xb1\xda\xca\xde\n"
        b"\x8ds\xaf\xc3\xba\xca\xb6\xde\xd7\xc5\xb8\xc3\xd3\xb2\x81R\xc4\xb2\xcb\n"
        b"\x96k\xc6\xb9\xdd\xb8\x83\x8e\xd4\xbf\xbc\xae\xb3\xb6\xde\xb1\xda\xca\xde\n"
        b"\xc2\xcf\xd7\xc5\xb2\xb6\xd7\xd4\xd2\xdb\xc4\xb2\xcb\n"
        b"\xcb\xc3\xde\xd8\xc9\xc4\xb7\xca\xc5\xd0\xc0\xde\xa6\xc5\xb6\xde\xbc\n"
        b"\xbb\xd1\xbb\xc9\xc5\xc2\xca\xb5\xdb\xb5\xdb\xb1\xd9\xb7\n"
        b"\xd0\xdd\xc5\xc6\xc3\xde\xb8\xc9\xce\xde\x81[\xc4\xd6\xca\xde\xda\n"
        b"\xce\xd2\xd7\xda\xd3\xbe\xbd\xde\n"
        b"\xb8\xc6\xd3\xbb\xda\xbd\xde\n"
        b"\xbb\xb3\xb2\xcc\xd3\xc9\xc6\n"
        b"\xdc\xc0\xbc\xca\xc5\xd8\xc0\xb2"
    )
    text = (
        "雨ニモマケズ\n風ニモマケズ\n雪ニモ夏ノ暑サニモマケヌ\n丈夫ナカラダヲモチ\n慾ハナク\n決シテ瞋ラズ\nイツモシヅカニワラッテヰル\n"
        "一日ニ玄米四合ト\n味噌ト少シノ野菜ヲタベ\nアラユルコトヲ\nジブンヲカンジョウニ入レズニ\nヨクミキキシワカリ\nソシテワスレズ\n野原ﾉ松ﾉ林ﾉ陰ﾉ\n"
        "小ｻﾅ萱ﾌﾞｷﾉ小屋ﾆヰﾃ\n東ﾆ病気ﾉｺﾄﾞﾓｱﾚﾊﾞ\n行ｯﾃ看病ｼﾃﾔﾘ\n西ﾆﾂｶﾚﾀ母ｱﾚﾊﾞ\n行ｯﾃｿﾉ稲ﾉ束ｦ負ﾋ\n南ﾆ死ﾆｻｳﾅ人ｱﾚﾊﾞ\n行ｯﾃｺﾊｶﾞﾗﾅｸﾃﾓｲヽﾄｲﾋ\n"
        "北ﾆｹﾝｸヮﾔｿｼｮｳｶﾞｱﾚﾊﾞ\nﾂﾏﾗﾅｲｶﾗﾔﾒﾛﾄｲﾋ\nﾋﾃﾞﾘﾉﾄｷﾊﾅﾐﾀﾞｦﾅｶﾞｼ\nｻﾑｻﾉﾅﾂﾊｵﾛｵﾛｱﾙｷ\nﾐﾝﾅﾆﾃﾞｸﾉﾎﾞーﾄﾖﾊﾞﾚ\nﾎﾒﾗﾚﾓｾｽﾞ\nｸﾆﾓｻﾚｽﾞ\nｻｳｲﾌﾓﾉﾆ\nﾜﾀｼﾊﾅﾘﾀｲ"
    )
    mocker.patch("pathlib.Path.read_bytes", return_value=cp932_text)

    assert normalize_bytes_text(cp932_text) == text
    assert read_text_normalized(pathlib.Path()) == text



import datetime
import pathlib
from pathlib import Path
from typing import Any, NamedTuple, cast

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.models import (
    CommandVerification,
    ConstVerification,
    FileResult,
    ResultStatus,
    VerificationFile,
    VerificationInput,
    VerificationResult,
    VerifyCommandResult,
)
from competitive_verifier.verify.verifier import (
    InputContainer,
    SplitState,
    Verifier,
    _now,  # pyright: ignore[reportPrivateUsage]
)

SUCCESS = ResultStatus.SUCCESS


def test_now():
    assert _now().tzinfo is not None


def test_get_file_timestamp_git_timestamp(mocker: MockerFixture):
    verifier = Verifier(
        VerificationInput(),
        timeout=1,
        default_tle=None,
        default_mle=None,
        prev_result=None,
        split_state=None,
        use_git_timestamp=True,
    )
    mocker.patch.dict(
        verifier.verifications.transitive_depends_on,
        {pathlib.Path("foo"): {pathlib.Path("foo"), pathlib.Path("bar")}},
    )

    get_commit_time = mocker.patch(
        "competitive_verifier.git.get_commit_time",
        return_value=datetime.datetime.max,
    )

    assert verifier.get_file_timestamp(pathlib.Path("foo")) == datetime.datetime.max
    get_commit_time.assert_called_once_with({pathlib.Path("foo"), pathlib.Path("bar")})


def test_get_file_timestamp_local(mocker: MockerFixture):
    class MockPath(NamedTuple):
        name: str
        time: datetime.datetime

        @property
        def st_mtime(self) -> float:
            assert self.time.tzinfo == datetime.timezone.utc
            return self.time.timestamp()

        def stat(self):
            return self

    verifier = Verifier(
        VerificationInput(),
        timeout=1,
        default_tle=None,
        default_mle=None,
        prev_result=None,
        split_state=None,
        use_git_timestamp=False,
    )

    foo_path = MockPath(
        "foo",
        datetime.datetime.fromisoformat("2001-02-03T04:05:05.789789+00:00"),
    )
    bar_path = MockPath(
        "bar",
        datetime.datetime.fromisoformat("2001-02-03T04:05:06.789789+00:00"),
    )

    mocker.patch(
        "competitive_verifier.verify.verifier._now",
        return_value=datetime.datetime.fromisoformat(
            "2010-08-31T15:26:56.456797+06:30"
        ),
    )
    mocker.patch.dict(
        verifier.verifications.transitive_depends_on,
        {foo_path: {foo_path, bar_path}},
    )

    get_commit_time = mocker.patch("competitive_verifier.git.get_commit_time")

    assert verifier.get_file_timestamp(
        cast("pathlib.Path", foo_path)
    ) == datetime.datetime.fromisoformat("2001-02-03T10:35:06+06:30")
    get_commit_time.assert_not_called()


class MockInputContainer(InputContainer):
    def __init__(
        self,
        obj: Any = None,
        *,
        prev_result: VerifyCommandResult | None = None,
        verification_time: datetime.datetime | None = None,
        file_timestamps: dict[Path, datetime.datetime] | None = None,
        split_state: SplitState | None = None,
    ) -> None:
        super().__init__(
            verifications=VerificationInput.model_validate(obj)
            if obj
            else VerificationInput(),
            verification_time=verification_time or datetime.datetime.now(),
            prev_result=prev_result,
            split_state=split_state,
        )

        self.file_timestamps = file_timestamps or {}

    def get_file_timestamp(self, path: Path) -> datetime.datetime:
        assert self.file_timestamps is not None
        dt = self.file_timestamps.get(path)
        assert dt is not None
        return dt


test_verification_files_params: list[
    tuple[InputContainer, dict[Path, VerificationFile]]
] = [
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {},
                    "bar": {},
                    "baz": {},
                },
            }
        ),
        {},
    ),
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "bar": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "baz": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                },
            }
        ),
        {
            Path("foo"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
            Path("bar"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
            Path("baz"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
        },
    ),
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "bar": {},
                    "baz": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                },
            }
        ),
        {
            Path("foo"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
            Path("baz"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
        },
    ),
]


@pytest.mark.parametrize(
    ("resolver", "expected"),
    test_verification_files_params,
)
def test_verification_files(
    resolver: InputContainer,
    expected: dict[Path, VerificationFile],
):
    assert resolver.verification_files == expected


test_file_need_verification_params: list[
    tuple[InputContainer, Path, FileResult, bool]
] = [
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2015, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        False,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2017, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        True,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2015, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.FAILURE,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        True,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2015, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SKIPPED,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        True,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2015, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        False,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2015, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2017, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        False,
    ),
]


test_remaining_verification_files_params: list[
    tuple[InputContainer, dict[Path, VerificationFile]]
] = [
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "bar": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "baz": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                },
            }
        ),
        {},
    ),
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {"verification": {"type": "command", "command": "true"}},
                    "bar": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "baz": {"verification": {"type": "command", "command": "true"}},
                },
            }
        ),
        {
            Path("foo"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("baz"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {"verification": {"type": "command", "command": "true"}},
                    "bar": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "baz": {"verification": {"type": "command", "command": "true"}},
                },
            },
            verification_time=datetime.datetime(9999, 5, 22),
            prev_result=VerifyCommandResult(
                total_seconds=1.5,
                files={
                    Path("baz"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1.5,
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime.datetime(2018, 7, 22),
                            )
                        ]
                    )
                },
            ),
        ),
        {
            Path("foo"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
]


@pytest.mark.parametrize(
    ("resolver", "expected"),
    test_remaining_verification_files_params,
)
def test_remaining_verification_files(
    resolver: InputContainer,
    expected: dict[Path, VerificationFile],
):
    assert resolver.remaining_verification_files == expected


test_current_verification_files_params: list[
    tuple[int, dict[Path, VerificationFile]]
] = [
    (
        0,
        {
            Path("bar/0.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("bar/1.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("bar/2.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        1,
        {
            Path("bar/3.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("baz/0.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("baz/1.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        2,
        {
            Path("baz/2.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("baz/3.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("foo/0.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        3,
        {
            Path("foo/1.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("foo/2.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("foo/3.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        4,
        {},
    ),
]


@pytest.mark.parametrize(
    ("resolver", "path", "file_result", "expected"),
    test_file_need_verification_params,
)
def test_file_need_verification(
    resolver: InputContainer,
    path: Path,
    file_result: FileResult,
    expected: bool,
    mocker: MockerFixture,
):
    mocker.patch.object(pathlib.Path, "exists", return_value=True)
    assert resolver.file_need_verification(path, file_result) == expected

    mocker.patch.object(pathlib.Path, "exists", return_value=False)
    assert not resolver.file_need_verification(path, file_result)


@pytest.mark.parametrize(
    ("index", "expected"),
    test_current_verification_files_params,
)
def test_current_verification_files(index: int, expected: dict[Path, VerificationFile]):
    command_verification = {"verification": {"type": "command", "command": "true"}}
    resolver = MockInputContainer(
        {
            "files": {
                "dummy/0.py": {},
                "foo/0.py": command_verification,
                "bar/0.py": command_verification,
                "baz/0.py": command_verification,
                "dummy/1.py": {},
                "foo/1.py": command_verification,
                "bar/1.py": command_verification,
                "baz/1.py": command_verification,
                "foo/2.py": command_verification,
                "dummy/2.py": {},
                "bar/2.py": command_verification,
                "baz/2.py": command_verification,
                "baz/3.py": command_verification,
                "dummy/3.py": {},
                "bar/3.py": command_verification,
                "foo/3.py": command_verification,
            },
        },
        split_state=SplitState(size=4, index=index),
    )
    remaining_verification_files = {
        Path("foo/0.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("bar/0.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("baz/0.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("foo/1.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("bar/1.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("baz/1.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("foo/2.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("bar/2.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("baz/2.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("foo/3.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("bar/3.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("baz/3.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
    }
    assert resolver.remaining_verification_files == remaining_verification_files
    assert resolver.current_verification_files == expected



from typing import Literal

from pydantic import Field

from competitive_verifier.models import ShellCommand, ShellCommandLike

from .base import OjVerifyUserDefinedConfig
from .user_defined import UserDefinedLanguage


class OjVerifyGoConfig(OjVerifyUserDefinedConfig):
    execute: ShellCommandLike = Field(
        default_factory=lambda: ShellCommand(
            command=["go", "run", "{basedir}/{path}"],
            env={"GO111MODULE": "off"},
        ),
    )


class GoLanguage(UserDefinedLanguage):
    extension: Literal["go"] = "go"  # pyright: ignore[reportIncompatibleVariableOverride]
    config: OjVerifyGoConfig = Field(default_factory=OjVerifyGoConfig)  # pyright: ignore[reportIncompatibleVariableOverride]



from datetime import datetime
from pathlib import Path
from typing import TypeVar

import pytest

from competitive_verifier.inout import merge
from competitive_verifier.models import (
    FileResult,
    ResultStatus,
    VerificationInput,
    VerificationResult,
    VerifyCommandResult,
)

inputs: list[tuple[list[VerificationInput], VerificationInput]] = [
    (
        [
            VerificationInput.model_validate(
                {
                    "files": {
                        "foo/bar.py": {},
                        "foo/baz.py": {
                            "path": "foo/baz.py",
                            "document_attributes": {
                                "title": "foo-baz",
                            },
                            "dependencies": ["foo/bar.py"],
                            "verification": [
                                {
                                    "type": "const",
                                    "status": "success",
                                }
                            ],
                        },
                    },
                }
            ),
        ],
        VerificationInput.model_validate(
            {
                "files": {
                    "foo/bar.py": {},
                    "foo/baz.py": {
                        "path": "foo/baz.py",
                        "document_attributes": {
                            "title": "foo-baz",
                        },
                        "dependencies": ["foo/bar.py"],
                        "verification": [
                            {
                                "type": "const",
                                "status": "success",
                            }
                        ],
                    },
                },
            }
        ),
    ),
    (
        [
            VerificationInput.model_validate(
                {
                    "files": {
                        "foo/bar.py": {},
                        "foo/baz.py": {
                            "path": "foo/baz.py",
                            "document_attributes": {
                                "title": "foo-baz",
                            },
                            "dependencies": ["foo/bar.py"],
                            "verification": [
                                {
                                    "type": "const",
                                    "status": "success",
                                }
                            ],
                        },
                    },
                }
            ),
            VerificationInput.model_validate(
                {
                    "files": {
                        "hoge/bar.py": {},
                        "hoge/baz.py": {
                            "path": "hoge/baz.py",
                            "document_attributes": {
                                "title": "foo-baz",
                            },
                            "dependencies": ["hoge/bar.py"],
                            "verification": [
                                {
                                    "type": "const",
                                    "status": "success",
                                }
                            ],
                        },
                    },
                }
            ),
        ],
        VerificationInput.model_validate(
            {
                "files": {
                    "foo/bar.py": {},
                    "foo/baz.py": {
                        "path": "foo/baz.py",
                        "document_attributes": {
                            "title": "foo-baz",
                        },
                        "dependencies": ["foo/bar.py"],
                        "verification": [
                            {
                                "type": "const",
                                "status": "success",
                            }
                        ],
                    },
                    "hoge/bar.py": {},
                    "hoge/baz.py": {
                        "path": "hoge/baz.py",
                        "document_attributes": {
                            "title": "foo-baz",
                        },
                        "dependencies": ["hoge/bar.py"],
                        "verification": [
                            {
                                "type": "const",
                                "status": "success",
                            }
                        ],
                    },
                },
            }
        ),
    ),
]

results: list[tuple[list[VerifyCommandResult], VerifyCommandResult]] = [
    (
        [
            VerifyCommandResult(
                total_seconds=2.5,
                files={
                    Path("foo"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1,
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime(2020, 2, 26, 23, 45),
                            )
                        ]
                    )
                },
            ),
        ],
        VerifyCommandResult(
            total_seconds=2.5,
            files={
                Path("foo"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime(2020, 2, 26, 23, 45),
                        )
                    ]
                )
            },
        ),
    ),
    (
        [
            VerifyCommandResult(
                total_seconds=2.5,
                files={
                    Path("foo"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1,
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime(2020, 2, 26, 23, 45),
                            )
                        ]
                    )
                },
            ),
            VerifyCommandResult(
                total_seconds=4.25,
                files={
                    Path("baz"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1,
                                status=ResultStatus.SKIPPED,
                                last_execution_time=datetime(2021, 2, 27, 23, 45),
                            )
                        ]
                    )
                },
            ),
            VerifyCommandResult(
                total_seconds=1,
                files={
                    Path("bar"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1,
                                status=ResultStatus.FAILURE,
                                last_execution_time=datetime(2020, 2, 27, 23, 45),
                            )
                        ]
                    )
                },
            ),
        ],
        VerifyCommandResult(
            total_seconds=7.75,
            files={
                Path("foo"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime(2020, 2, 26, 23, 45),
                        )
                    ]
                ),
                Path("bar"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.FAILURE,
                            last_execution_time=datetime(2020, 2, 27, 23, 45),
                        )
                    ]
                ),
                Path("baz"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime(2021, 2, 27, 23, 45),
                        )
                    ]
                ),
            },
        ),
    ),
]

T = TypeVar("T", VerificationInput, VerifyCommandResult)


@pytest.mark.parametrize(
    ("objects", "expected"),
    inputs + results,
)
def test_merge(objects: list[T], expected: T):
    assert merge(objects) == expected



import os
import pathlib

import pytest
from pytest_mock import MockerFixture
from pytest_mock.plugin import MockType

from competitive_verifier.download import download_files as download
from competitive_verifier.models import (
    ConstVerification,
    ProblemVerification,
    ResultStatus,
    VerificationFile,
)
from competitive_verifier.oj import problem


@pytest.fixture
def mock_problem(mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.chdir(pathlib.Path(__file__).parent)
    mocker.patch.dict(os.environ, {"YUKICODER_TOKEN": "YKTK"}, clear=True)

    return {
        problem.YukicoderProblem: mocker.patch.object(
            problem.YukicoderProblem,
            "download_system_cases",
            return_value=[],
        ),
        problem.LibraryCheckerProblem: mocker.patch.object(
            problem.LibraryCheckerProblem,
            "download_system_cases",
            return_value=[],
        ),
    }


@pytest.mark.parametrize(
    "url_or_file",
    [
        [
            "https://judge.yosupo.jp/problem/aplusb",
            "https://judge.yosupo.jp/problem/aplusb",
            "https://yukicoder.me/problems/no/1088",
        ],
        [
            VerificationFile(
                verification=ProblemVerification(
                    problem="https://judge.yosupo.jp/problem/aplusb",
                    command="true",
                )
            ),
            VerificationFile(
                verification=[
                    ProblemVerification(
                        problem="https://yukicoder.me/problems/no/1088",
                        command="true",
                    ),
                    ConstVerification(status=ResultStatus.FAILURE),
                ]
            ),
        ],
    ],
)
def test_oj_download(
    url_or_file: str | VerificationFile | list[str | VerificationFile],
    mock_problem: dict[type[problem.Problem], MockType],
):
    download(url_or_file=url_or_file)

    assert set(mock_problem.keys()) == {
        problem.LibraryCheckerProblem,
        problem.YukicoderProblem,
    }

    mock_library_checker = mock_problem[problem.LibraryCheckerProblem]
    mock_library_checker.assert_called_once_with()

    mock_yuki_coder = mock_problem[problem.YukicoderProblem]
    mock_yuki_coder.assert_called_once_with()



import os

import pytest
from pytest_mock import MockerFixture

from competitive_verifier import github


@pytest.fixture
def workflow_ref(mocker: MockerFixture, request: pytest.FixtureRequest):
    mocker.patch.dict(
        os.environ,
        {"GITHUB_WORKFLOW_REF": request.param} if request.param else {},
        clear=True,
    )


@pytest.mark.usefixtures("workflow_ref")
@pytest.mark.parametrize(
    ("workflow_ref", "expected"),
    [
        (
            None,
            None,
        ),
        (
            "octocat/hello-world/.github/workflows/my-workflow.yml@refs/heads/my_branc",
            "my-workflow.yml",
        ),
    ],
    indirect=["workflow_ref"],
)
def test_workflow_filename(expected: str | None):
    assert github.env.get_workflow_filename() == expected


def test_github_envvar(mocker: MockerFixture):
    mocker.patch.dict(
        os.environ,
        {},
        clear=True,
    )
    assert github.env.get_ref_name() is None
    assert github.env.get_repository() is None
    assert github.env.get_workflow_name() is None

    mocker.patch.dict(
        os.environ,
        {
            "GITHUB_REF_NAME": "qwerty",
            "GITHUB_REPOSITORY": "asdfgh",
            "GITHUB_WORKFLOW": "zxcvbn",
        },
        clear=True,
    )
    assert github.env.get_ref_name() == "qwerty"
    assert github.env.get_repository() == "asdfgh"
    assert github.env.get_workflow_name() == "zxcvbn"



# Python Version: 3.x
import abc
import pathlib
from collections.abc import Sequence
from typing import Any

from pydantic import BaseModel, ConfigDict

from competitive_verifier.models import ShellCommandLike
from competitive_verifier.oj.verify.languages import special_comments


class LanguageEnvironment(abc.ABC):
    @property
    @abc.abstractmethod
    def name(self) -> str: ...

    @abc.abstractmethod
    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> ShellCommandLike | None: ...

    @abc.abstractmethod
    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> ShellCommandLike: ...


class Language(BaseModel):
    def list_attributes(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> dict[str, Any]:
        attributes: dict[str, Any] = dict(special_comments.list_special_comments(path))
        attributes.setdefault("links", [])
        attributes["links"].extend(special_comments.list_embedded_urls(path))
        return attributes

    @abc.abstractmethod
    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]: ...

    def bundle(self, path: pathlib.Path, *, basedir: pathlib.Path) -> bytes | None:
        return None

    @abc.abstractmethod
    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> Sequence[LanguageEnvironment]: ...


class OjVerifyLanguageConfig(BaseModel):
    model_config = ConfigDict(extra="allow")


class OjVerifyUserDefinedConfig(OjVerifyLanguageConfig):
    execute: ShellCommandLike
    compile: ShellCommandLike | None = None
    bundle: ShellCommandLike | None = None
    list_attributes: ShellCommandLike | None = None
    list_dependencies: ShellCommandLike | None = None



import pathlib
from argparse import ArgumentParser
from collections import Counter
from collections.abc import Iterable
from functools import reduce
from logging import getLogger
from typing import Literal, TypeVar

from pydantic import Field

from competitive_verifier.arg import (
    ResultJsonArguments,
    VerboseArguments,
    WriteSummaryArguments,
)
from competitive_verifier.models import (
    ResultStatus,
    VerificationInput,
    VerifyCommandResult,
)

logger = getLogger(__name__)

T = TypeVar("T", VerificationInput, VerifyCommandResult)


def merge(results: Iterable[T]) -> T:
    return reduce(lambda a, b: a.merge(b), results)


class MergeInput(VerboseArguments):
    subcommand: Literal["merge-input"] = Field(
        default="merge-input",
        description="Merge verify_files.json`",
    )
    verify_files_json: list[pathlib.Path]

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        parser.add_argument(
            "verify_files_json",
            nargs="+",
            help="verify_files.json files",
            type=pathlib.Path,
        )

    def _run(self) -> bool:
        result = merge(
            map(VerificationInput.parse_file_relative, self.verify_files_json)
        )
        print(result.model_dump_json(exclude_none=True))
        return True


class MergeResult(WriteSummaryArguments, ResultJsonArguments, VerboseArguments):
    subcommand: Literal["merge-result"] = Field(
        default="merge-result",
        description="Merge result of `verify`",
    )

    def merge(self) -> VerifyCommandResult:
        return merge(map(VerifyCommandResult.parse_file_relative, self.result_json))

    def _run(self) -> bool:
        result = self.merge()
        self.write_result(result)

        print(result.model_dump_json(exclude_none=True))
        return True


class Check(ResultJsonArguments, VerboseArguments):
    subcommand: Literal["check"] = Field(
        default="check",
        description="Check result of `verify`",
    )

    def _run(self) -> bool:
        result = merge(map(VerifyCommandResult.parse_file_relative, self.result_json))

        counter = Counter(
            r.status for fr in result.files.values() for r in fr.verifications
        )

        for st in ResultStatus:
            print(f"{st.value}: {counter.get(st, 0)}")

        if counter[ResultStatus.FAILURE] > 0:
            if self.verbose:
                logger.error("Failure test count: %d", counter[ResultStatus.FAILURE])
            return False
        return True



from argparse import ArgumentParser
from collections.abc import Iterable
from itertools import chain
from logging import getLogger
from typing import Literal

from pydantic import Field

from competitive_verifier import oj
from competitive_verifier.arg import (
    OptionalVerifyFilesJsonArguments,
    VerboseArguments,
)
from competitive_verifier.models import (
    ProblemVerification,
    VerificationFile,
    VerificationInput,
)
from competitive_verifier.resource import try_ulimit_stack

logger = getLogger(__name__)

UrlOrVerificationFile = str | VerificationFile


def _parse_single_url(url_or_file: UrlOrVerificationFile) -> Iterable[str]:
    if isinstance(url_or_file, str):
        yield url_or_file
    else:
        for v in url_or_file.verification_list:
            if isinstance(v, ProblemVerification):
                yield v.problem


def parse_urls(
    url_or_file: UrlOrVerificationFile | Iterable[UrlOrVerificationFile],
) -> set[str]:
    if isinstance(url_or_file, (str, VerificationFile)):
        return set(_parse_single_url(url_or_file))

    return set(chain.from_iterable(map(_parse_single_url, url_or_file)))


def download_files(
    url_or_file: UrlOrVerificationFile | Iterable[UrlOrVerificationFile],
    *,
    check: bool = False,
    group_log: bool = False,
) -> bool:
    result = True
    try_ulimit_stack()

    for url in parse_urls(url_or_file):
        if not oj.download(url, group_log=group_log):
            result = False

    if check and not result:
        raise RuntimeError("Failed to download")
    return result


class Download(OptionalVerifyFilesJsonArguments, VerboseArguments):
    subcommand: Literal["download"] = Field(
        default="download",
        description="Download problems",
    )
    urls: list[str] = Field(default_factory=list)

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        parser.add_argument(
            "urls",
            nargs="*",
            help="A list of problem URL",
        )

    def _run(self) -> bool:
        logger.debug("arguments:%s", self)
        logger.info(
            "verify_files_json=%s, urls=%s",
            str(self.verify_files_json),
            self.urls,
        )
        files: list[VerificationFile] = []
        if self.verify_files_json:
            files.extend(
                VerificationInput.parse_file_relative(
                    self.verify_files_json
                ).files.values()
            )

        return download_files(files + self.urls, group_log=True)



import json
import pathlib
from typing import Any

import pytest

from competitive_verifier import app

from .data.integration_data import IntegrationData
from .types import ConfigDirSetter, FilePaths


class CompileFailureData(IntegrationData):
    @classmethod
    def input_name(cls) -> str:
        return "CompileFailure"

    def expected_verify_result(self) -> dict[str, Any]:
        return {
            "total_seconds": 1312.56,
            "files": {
                "verify.json": {
                    "newest": True,
                    "verifications": [
                        {
                            "verification_name": "Compile Failure",
                            "elapsed": 8351.0,
                            "last_execution_time": "1995-04-15T15:40:58.150000+03:00",
                            "status": "failure",
                        },
                        {
                            "verification_name": "Command Failure",
                            "elapsed": 1010.0,
                            "heaviest": 40.0,
                            "last_execution_time": "1989-12-16T21:45:15.390000+02:00",
                            "slowest": 101.0,
                            "status": "failure",
                            "testcases": [
                                {
                                    "elapsed": 5.59,
                                    "memory": 33.3,
                                    "name": "example_00",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 8.91,
                                    "memory": 23.16,
                                    "name": "example_01",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 5.67,
                                    "memory": 10.76,
                                    "name": "random_00",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 3.37,
                                    "memory": 92.99,
                                    "name": "random_01",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 0.81,
                                    "memory": 69.34,
                                    "name": "random_02",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 8.9,
                                    "memory": 99.46,
                                    "name": "random_03",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 9.87,
                                    "memory": 22.2,
                                    "name": "random_04",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 6.55,
                                    "memory": 53.74,
                                    "name": "random_05",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 9.07,
                                    "memory": 51.0,
                                    "name": "random_06",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 2.73,
                                    "memory": 40.89,
                                    "name": "random_07",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 4.05,
                                    "memory": 54.04,
                                    "name": "random_08",
                                    "status": "RE",
                                },
                                {
                                    "elapsed": 1.0,
                                    "memory": 62.53,
                                    "name": "random_09",
                                    "status": "RE",
                                },
                            ],
                        },
                    ],
                }
            },
        }

    def expected_verify_json(self) -> dict[str, Any]: ...


@pytest.fixture
def compile_failure_integration_data(
    monkeypatch: pytest.MonkeyPatch,
    file_paths: FilePaths,
    set_config_dir: ConfigDirSetter,
) -> IntegrationData:
    return CompileFailureData(monkeypatch, set_config_dir, file_paths)


@pytest.mark.integration
@pytest.mark.order(-500)
@pytest.mark.usefixtures("additional_path", "mock_verification")
def test_compile_failure(
    compile_failure_integration_data: IntegrationData,
):
    verify = compile_failure_integration_data.targets_path / "verify.json"
    result = compile_failure_integration_data.config_dir_path / "result.json"

    parsed = app.ArgumentParser().parse(
        ["verify", "--verify-json", str(verify), "--output", str(result)]
    )
    assert isinstance(parsed, app.Verify)
    assert parsed.run()

    assert (
        json.loads(pathlib.Path(result).read_bytes())
        == compile_failure_integration_data.expected_verify_result()
    )
